####################################################################################################
# cortex.py
# Simple tools for dealing with the cortical surface in Python
# By Noah C. Benson

import numpy as np
import scipy as sp
import scipy.spatial as space
import neuropythy.geometry as geo
from neuropythy.immutable import Immutable
import nibabel.freesurfer.io as fsio
import os
import itertools
import collections
from pysistence import make_dict
import pysistence

class CorticalMesh(Immutable):
    '''CorticalMesh is a class that handles properties of the cortical surface mesh. 
       The cortical mesh representation contains only three pieces of concrete data and many pieces
       of derived data. All derived data is lazily computed when requested then saved. Any changes
       to the concrete data results in the relevant derived data being cleared (and thus 
       recalculated on next request). Note that this will not work if you simply modify the numpy
       arrays, such as mesh.coordinates[:,10] = [0,0,0]; in this case, the function mesh.flush()
       or mesh.flush('coordinates') should be called to make sure that all cache is flushed.

       The concrete data members of CorticalMesh are:
         * coordinates: the x,y,z coordinates of the vertices; this matrix must be 2 or 3 by n
           where n is the number of vertices in the mesh.
         * faces: the integer triples that form the triangles of the mesh
         * options: a dictionary of all options to CorticalMesh; note that one of these options is
           \'meta_data\', which may be anything, but should usually be a dictionary in which user-
           defined data may be placed.

       The derived (lazy) data members of CorticalMesh are:
         * edges: a 2 x m numpy array of the edges in the mesh
         * face_angles: a 3 x q vector of the face angles
         * face_normals: a 3 x q vector of the normal vectors to each face

       CorticalMesh supports the following methods:
         * vertex_edges, vertex_faces, edge_faces: these functions, when given a vertex, vertex, or
           edge, respectively, yield a list of the edges, faces, or faces that the given item is a
           part of; for example, mesh.vertex_edges(u) yields the list of edges containing u.
         * select: mesh.select(f) can be used to select only those vertices from the mesh such that
           f(v) yields true; a new mesh object is returned.
         * add_property, remove_property, property_value, prop: these methods deal with the mesh's
           properties.
         * map_vertices: this function maps over each vertex in the mesh and provides the given
           function with a dictionary of the properties, coordinate, and label of each vertex.'''


    ################################################################################################
    # Nested Classes

    class Index:
        def __init__(self, vidx, eidx, fidx):
            self.vertex_index = vidx
            self.edge_index = eidx
            self.face_index = fidx
        def __repr__(self):
            return "CorticalMesh.Index(<%d vertices>)" % len(self.vertex_index)
        def __getitem__(self, index):
            if isinstance(index, tuple):
                if len(index) == 3:   return self.face_index.get(index, None)
                elif len(index) == 2: return self.edge_index.get(index, None)
                elif len(index) == 1: return self.vertex_index.get(index[0], None)
                else:                 raise ValueError('Unrecognized mesh item: %s' % index)
            elif isinstance(index, (list, set)):
                if len(index) > 3 or not all(isinstance(i,int) for i in index):
                    if isinstance(index, list): return [self[x]    for x in index]
                    else:                       return {x: self[x] for x in index}
                else:
                    return [
                        lambda idx: [],
                        lambda idx: self.vertex_index.get(idx[0], None),
                        lambda idx: self.edge_index.get(tuple(idx), None),
                        lambda idx: self.face_index.get(tuple(idx), None)
                        ][len(index)](index)
            elif isinstance(index, np.ndarray):
                return self[index.tolist()]
            elif isinstance(index, int):
                return self.vertex_index.get(index, None)
            else:
                raise ValueError('Unrecognized mesh item: %s' % index)

                

    ################################################################################################
    # Static Functions
    # These are mostly for calculating instance members lazily from other members

    @staticmethod
    def calculate_vertex_data(faces, edges, coords, vlab):
        n = coords.shape[1]
        vertex_idx = {v:i for (i,v) in enumerate(vlab)}
        d = collections.defaultdict(list)
        for idx in itertools.product(*map(range, edges.shape)):
            d[edges[idx]].append(idx)
        edge_idx = [map(lambda q: q[1], d[i]) for i in range(n)]
        d = collections.defaultdict(list)
        for idx in itertools.product(*map(range, faces.shape)):
            d[faces[idx]].append(idx)
        face_idx = [map(lambda q: q[1], d[i]) for i in range(n)]
        return (vertex_idx, edge_idx, face_idx)

    @staticmethod
    def calculate_edge_data(faces):
        limit = max(faces.flatten()) + 1
        edge2face = {}#sp.sparse.lil_matrix((limit, limit), dtype='int')
        idx = {}#sp.sparse.lil_matrix((limit, limit), dtype='int')
        edge_list = [None for i in range(3*faces.size)]
        k = 0
        rng = range(faces.shape[1])
        for (e,i) in zip(
            zip(np.concatenate((faces[0], faces[1], faces[2])),
                np.concatenate((faces[1], faces[2], faces[0]))),
            np.concatenate((rng, rng, rng))):
            if e not in idx:
                idx[e] = k
                idx[e[::-1]] = k
                edge_list[k] = e
                edge2face[e] = i
                k += 1
            elif e in edge2face:
                edge2face[e[::-1]] = i
        for ((a,b),i) in edge2face.items():
            edge2face[(b,a)] = i
        return (np.array(edge_list[0:k]).transpose(), idx, edge2face)

    @staticmethod
    def calculate_face_data(faces):
        idx = {}
        for (a,b,c,i) in zip(faces[0], faces[1], faces[2], range(faces.shape[1])):
            idx[(a,b,c)] = i
            idx[(b,c,a)] = i
            idx[(c,b,a)] = i
            idx[(a,c,b)] = i
            idx[(b,a,c)] = i
            idx[(c,a,b)] = i
        return idx

    @staticmethod
    def calculate_face_angles(faces, coords):
        X = np.array([coords[:,faces[0]], coords[:,faces[1]], coords[:,faces[2]]])
        sides = [X[1] - X[0], X[2] - X[1], X[0] - X[2]]
        normed_sides = map(
            lambda side: side / np.repeat([np.sqrt((side**2).sum(0))], coords.shape[0], 0),
            sides)
        dps = [(normed_sides[0] * (-normed_sides[2])).sum(0),
               (normed_sides[1] * (-normed_sides[0])).sum(0),
               (normed_sides[2] * (-normed_sides[1])).sum(0)]
        return np.arccos(dps)

    @staticmethod
    def calculate_face_normals(faces, coords):
        if coords.shape[0] != 3:
            raise ValueError('Face normals cannot be computed for 2D meshes')
        X = np.array([coords[:,faces[0]], coords[:,faces[1]], coords[:,faces[2]]])
        s1 = X[1] - X[0]
        s2 = X[2] - X[0]
        crosses = np.array([s1[1]*s2[2] - s1[2]*s2[1],
                            s1[2]*s2[0] - s1[0]*s2[2],
                            s1[0]*s2[1] - s1[1]*s2[0]])
        norms = np.sqrt((crosses**2).sum(0))
        w = (norms == 0)
        norms[w] = 1.0
        return np.where(w, 0, crosses / np.repeat([norms], 3, 0))

    @staticmethod
    def calculate_vertex_normals(vfaces, fnorms):
        tmp = np.array(
            map(
                lambda fs: fnorms[:,fs].sum(1),
                vfaces)).transpose()
        norms = np.sqrt((tmp ** 2).sum(0))
        w = (norms == 0)
        norms[w] = 1.0
        return np.where(w, 0, tmp / np.repeat([norms], 3, 0))

    @staticmethod
    def calculate_index(vertex_index, edge_index, face_index):
        return CorticalMesh.Index(vertex_index, edge_index, face_index)

    @staticmethod
    def calculate_spherical_coordinates(coords):
        n = coords.shape[1]
        centroid = np.mean(coords, 1)
        coords = np.array(
            [coords[0] - centroid[0],
             coords[1] - centroid[1],
             coords[2] - centroid[2]])
        rho = np.sqrt((coords ** 2).sum(0))
        normed = coords / np.array([rho]).repeat(3, 0)
        theta = np.arctan2(normed[0], normed[1])
        phi = np.arcsin(normed[2])
        return np.array([theta, phi, rho])


    ################################################################################################
    # Lazy/Static Interface
    # This code handles the lazy interface; only three data are non-lazy and these are coordinates,
    # faces, and options; they may be set directly.

    # These functions and the following static variable allow one to set particular members of the
    # object directly:
    @staticmethod
    def _check_coordinates(self, val):
        # Coordinates must be stored as a np array
        if type(val) is np.ndarray: x = val
        elif type(val) is list: x = np.array(val)
        else: raise ValueError('coordinates must be a matrix of 2D or 3D points')
            # Coordinates must be 2D or 3D
        if len(x.shape) != 2:
            raise ValueError('coordinates must be a matrix of 2D or 3D points')
        elif x.shape[1] == 2 or x.shape[1] == 3:
            x = x.transpose()
        elif x.shape[0] != 2 and x.shape[0] != 3:
            raise ValueError('coordinates must be a matrix of 2D or 3D points')
        if not x.flags['WRITEABLE']:
            x = x.copy()
            x.setflags(write=False)
        return x
    @staticmethod
    def _check_faces(self, val):
        # Faces must be an array
        if type(val) is np.ndarray: x = val
        elif type(val) is list:
            x = np.array(val)
        else:
            raise ValueError('faces must be an integer metrix of faces')
        # Faces must be an integer array
        if not issubclass(x.dtype.type, np.integer):
            raise ValueError('faces must be an integer metrix of faces')
        # Faces must be n x 3 or 3 x n
        if x.shape[1] == 3:
            x = x.transpose()
        elif x.shape[0] != 3:
            raise ValueError('faces must be an integer metrix of faces')
        if not x.flags['WRITEABLE']:
            x = x.copy()
            x.setflags(write=False)
        return x
    @staticmethod
    def _check_vertex_labels(self, val):
        # Must be a 1D list
        if type(val) is np.ndarray: x = val
        elif type(val) is list: x = np.array(val)
        else: raise ValueError('vertex_label must be an integer list of vertex labels')
        if len(x.shape) != 1 or not issubclass(x.dtype.type, np.integer):
            raise ValueError('vertex_label must be an integer list of vertex labels')
        if x.shape[0] != self.coordinates.shape[1]:
            raise ValueError('vertex_label must match coordinates in size')
        if not x.flags['WRITEABLE']:
            x = x.copy()
            x.setflags(write=False)
        return x
    @staticmethod
    def _check_options(self, val):
        # Options just have to be a dictionary and are converted to an immutable one
        if not isinstance(val, dict):
            raise ValueError('options must be a dictionary')
        if type(val) is pysistence.persistent_dict.PDict:
            return val
        else:
            return make_dict(**val)
    @staticmethod
    def _check_property(self, name, val):
        if not isinstance(val, np.ndarray) and not isinstance(val, list):
            raise ValueError('property values must be lists or numpy arrays')
        elif len(val) != len(self.vertex_labels):
            raise ValueError('property values must be of equal length to vertex count')
        return True
    @staticmethod
    def _check_properties(self, val):
        if not isinstance(val, dict):
            raise ValueError('properties must be a dictionary')
        n = len(self.vertex_labels)
        for (k, v) in val.iteritems(): CorticalMesh._check_property(self, k, v)
        if type(val) is pysistence.persistent_dict.PDict:
            return val
        else:
            return make_dict(**val)
    __settable_members = {
        'coordinates': lambda m,v: CorticalMesh._check_coordinates(m,v),
        'faces': lambda m,v: CorticalMesh._check_faces(m,v),
        'vertex_labels': lambda m,v: CorticalMesh._check_vertex_labels(m,v),
        'properties': lambda m,v: CorticalMesh._check_properties(m,v),
        'options': lambda m,v: CorticalMesh._check_options(m,v)}

    # This static variable explains the dependency hierarchy in cached data
    __lazy_members = {
        'vertex_data': (
            ('faces', 'edges', 'coordinates', 'vertex_labels'),
            lambda F,E,X,L: CorticalMesh.calculate_vertex_data(F, E, X, L)),
        'vertex_index': (('vertex_data',), lambda VD: VD[0]),
        'vertex_edge_index': (('vertex_data',), lambda VD: VD[1]),
        'vertex_face_index': (('vertex_data',), lambda VD: VD[2]),

        'edge_data': (('faces',), lambda F: CorticalMesh.calculate_edge_data(F)),
        'edges': (('edge_data',), lambda ED: ED[0]),
        'edge_index': (('edge_data',), lambda ED: ED[1]),
        'indexed_edges': (('edges','vertex_index'), 
                          lambda E,VI: np.asarray([[VI[u] for u in E[0]], [VI[v] for v in E[1]]])),
        'edge_face_index': (('edge_data',), lambda ED: ED[2]),

        'face_index': (('faces',), lambda F: CorticalMesh.calculate_face_data(F)),
        'indexed_faces': (('faces','vertex_index'), 
                          lambda F,VI: np.asarray([[VI[a] for a in F[0]],
                                                   [VI[b] for b in F[1]],
                                                   [VI[c] for c in F[2]]])),

        'index': (
            ('vertex_index', 'edge_index', 'face_index'),
            lambda VI, EI, FI: CorticalMesh.calculate_index(VI, EI, FI)),

        'edge_coordinates': (('indexed_edges', 'coordinates'), 
                             lambda E,X: np.asarray([X[:,E[0]], X[:, E[1]]])),
        'face_coordinates': (('indexed_faces', 'coordinates'),
                             lambda F,X: np.asarray([X[:,F[0]], X[:, F[1]], X[:, F[2]]])),

        'edge_lengths': (('edge_coordinates',), 
                         lambda EX: np.sqrt(np.power(EX[0] - EX[1], 2).sum(0))),

        'face_angles': (
            ('indexed_faces', 'coordinates'),
            lambda F,X: CorticalMesh.calculate_face_angles(F,X)),
        'face_normals': (
            ('indexed_faces', 'coordinates'),
            lambda F,X: CorticalMesh.calculate_face_normals(F,X)),
        'vertex_normals': (
            ('vertex_face_index', 'face_normals'),
            lambda VF,FN: CorticalMesh.calculate_vertex_normals(VF, FN)),

        'vertex_spatial_hash': (('vertex_coordinates',), lambda X: space.cKDTree(X.T)),
        'face_spatial_hash': (('face_coordinates',), lambda FX: space.cKDTree(FX.mean(0).T)),

        'meta_data': (
             ('options',),
             lambda opts: opts.get('meta_data', {})),

        'spherical_coordinates': (
            ('coordinates',),
            lambda X: CorticalMesh.calculate_spherical_coordinates(X))}
        
   
    
        
    ################################################################################################
    # Constructor

    def __init__(self, coords, faces, **args):
        Immutable.__init__(self, CorticalMesh.__settable_members, {}, CorticalMesh.__lazy_members)
        # Setup coordinates
        self.coordinates = coords
        # And faces...
        self.faces = faces
        # If vertex labels were provided, make sure to set these
        self.vertex_labels = args.pop('vertex_labels', range(self.coordinates.shape[1]))
        # Same with properties
        self.properties = args.pop('properties', make_dict())
        # And with hemisphere and subject
        self.__dict__['hemisphere'] = args.pop('hemisphere', None)
        self.__dict__['subject'] = args.pop('subject', None)
        # Finally, set the remaining options...
        self.options = args


    ################################################################################################
    # String versions of the object

    def __repr__(self):
        return "CorticalMesh(<" + str(self.coordinates.shape[1]) + " vertices>, <" \
            + str(self.faces.shape[1]) + " faces>)"

    ################################################################################################
    # Methods

    def vertex_edges(self, U):
        '''mesh.vertex_edges(U) yields a list of the edges that contain vertex U. If U is a list of
           vertices, then this function will automatically thread over it.'''
        if isinstance(U, int):
            return self.vertex_edge_index[U]
        else:
            return map(lambda u: self.vertex_edges(u), U)
    def vertex_faces(self, U):
        '''mesh.vertex_faces(U) yields a list of the faces that contain vertex U. If U is a list of
           vertices, then this function will automatically thread over it.'''
        if isinstance(U, int):
            return self.vertex_face_index[U]
        else:
            return map(lambda u: self.vertex_faces(u), U)
    def edge_faces(self, E):
        '''mesh.edge_faces(E) yields a list of the faces that contain edge E. If E is a list of
           edges, then this function will automatically thread over it.'''
        if isinstance(E, tuple) and len(E) == 2:
            return self.edge_face_index[E]
        else:
            return map(lambda e: self.edge_faces(e), E)            

    def select(self, filt, filter_vertices=True, filter_edges=False, filter_faces=False):
        '''mesh.select(filt) yields a new CorticalMesh object that is identical to mesh except
           that only the vertices u for which filt(u) yields true will be retained. Three optional
           arguments, filter_vertices (true), filter_edges (false), and filter_faces (false) may
           also be specified. For any of these which is true, those elements will be filtered using
           the filt function. For any that is false, the elements will be excluded only if their
           subparts are excluded. For any that is neither true or false, it must be a function, in
           which case that function is used as a filter in place of the filt function.
           If filt is instead a list of vertex labels, then filter_vertices is not used and instead
           the given list is used as the initial vertex filter.'''
        vf = filt if filter_vertices is True \
            else filter_vertices if filter_vertices is not False \
            else None
        ef = filt if filter_edges is True \
            else filter_edges if filter_edges is not False \
            else None
        ff = filt if filter_faces is True \
            else filter_faces if filter_faces is not False \
            else None
        # Find the included vertices:
        if isinstance(vf, np.ndarray): vincl = vf
        elif isinstance(vf, list):     vincl = vf
        elif vf is not None:           vincl = [u for u in self.vertex_labels if vf(u)]
        else:                          vincl = self.vertex_labels
        vincl = set(self.index[vincl])
        # Find the included edges:
        es = self.indexed_edges
        edge_idcs = range(self.edges.shape[1])
        if isinstance(ef, (list, set)):
            ef = np.asarray(ef)
        if ef is None:
            eincl = edge_idcs
        elif isinstance(ef, np.ndarray): 
            if len(ef.shape) == 1:
                eincl = self.index[ef]
            else:
                eincl = self.index[ef.T if ef.shape[1] != 2 else ef]
        else:
            elst = self.edge_list.T
            if ef is None: eincl = edge_idcs
            else:          eincl = [e for e in edge_idcs if ef(tuple(elst[e]))]
        # filter by vertices
        eincl = set(map(int, eincl))
        eincl.intersection_update(
            np.union1d([], [l for idcs in self.vertex_edges(vincl)
                              for l in idcs
                              if es[0,l] in vincl and es[1,l] in vincl]))
        # Find the included faces
        face_idcs = range(self.faces.shape[1])
        fs = self.indexed_faces
        if isinstance(ff, (list, set)):
            ff = np.asarray(ff)
        if ff is None:
            fincl = face_idcs
        elif isinstance(ff, np.ndarray): 
            if len(ff.shape) == 1:
                fincl = ff.tolist()
            else:
                fincl = self.index[ff.T if ff.shape[1] != 3 else ff]
        else:
            if ff is None: fincl = face_idcs
            else:          fincl = [f for f in face_idcs if ff(tuple(self.faces[:,f]))]
        # filter by vertices
        fincl = set(map(int, fincl))
        fincl.intersection_update(
            np.union1d([], [l for idcs in self.vertex_faces(vincl)
                              for l in idcs
                              if fs[0,l] in vincl and fs[1,l] in vincl and fs[2,l] in vincl]))
        vincl = list(vincl)
        eincl = list(eincl)
        fincl = list(fincl)
        # Make the subsets
        I = self.index[vincl]
        X = self.coordinates[:, I]
        V = self.vertex_labels[I]
        F = self.faces[:, fincl]
        opts = self.options.copy()
        meta = opts.get('meta_data', {}).copy()
        opts = opts.without('meta_data')
        meta = meta.using(source_mesh=self)
        props = {name: prop[I] for (name,prop) in self.properties.iteritems()}
        return CorticalMesh(X, F, vertex_labels=V, meta_data=meta, properties=props, **opts)

    def add_property(self, name, prop=Ellipsis):
        '''mesh.add_property(name, prop) adds (or overwrites) the given property with the given name
           in the given mesh. The name must be a valid dictionary key and the prop argument must be
           a list of numpy array of values, one per vertex.
           mesh.add_property(d) adds all of the properties in the given dictionary of properties, 
           d.
           Note that in either case, if the value of prop or the value in a dictionary item is
           None, then the item is removed from the property list instead of being added.'''
        n = len(self.vertex_labels)
        if prop is Ellipsis:
            if isinstance(name, dict):
                for (n,p) in name.iteritems():
                    self.add_property(n, p)
            else:
                raise ValueError('add_property must be called with a name and propery or a dict')
        else:
            if prop is None:
                self.remove_property(name)
            else:
                CorticalMesh._check_property(self, name, prop)
                self.__dict__['properties'] = self.properties.using(**{name: prop})

    def has_property(self, name):
        '''mesh.has_property(name) yields True if the given mesh contains the property with the
           given name. Note that this checks the hemisphere properties as well as the mesh
           properties and should be used preferentially over (name in mesh.properties).'''
        if name in self.properties:
            return True
        elif self.hemisphere is not None:
            return self.hemisphere.has_property(name)
        else:
            return False

    def remove_property(self, name):
        '''mesh.remove_property(name) removes the property with the given name from the given mesh.
           The name argument may also be an iterable collection of names, in which case all are
           removed. Note that you cannot remove properties inherited from the Hemisphere object;
           these must be removed from the Hemisphere.'''
        if hasattr(name, '__iter__'):
            for n in name: self.remove_property(n)
        elif name in self.properties:
            self.__dict__['properties'] = self.properties.without(name)
    
    def property_value(self, name):
        '''mesh.property_value(name) yields a list of the property values with the given name. If
           name is an iterable, then property_value automatically threads across it. If no such
           property is found in the mesh, then None is returned.'''
        if hasattr(name, '__iter__'):
            return map(lambda n: self.property_value(n), name)
        elif name in self.properties:
            return self.properties[name]
        elif self.hemisphere is not None:
            return self.hemisphere.property_value(name)

    def property_names(self):
        '''mesh.property_names() yields a set of property names for the given mesh. Note that this
           lists properties inherited from the hemisphere class, so it should be used preferentially
           over mesh.properties.keys().'''
        if self.hemisphere is not None:
            return set(list(self.properties.keys()) + self.hemisphere.property_names())
        else:
            return set(self.properties.keys())
            
    def prop(self, name, arg=Ellipsis):
        '''mesh.prop(...) is a generic function for handling mesh properties. It may be called
           with a variety of arguments:
             * mesh.prop(name) is equivalent to mesh.property_value(name) if name is either not
               an iterable or is an iterable that is not a dictionary
             * mesh.prop(d) is equivalent to mesh.add_property(d) if d is a dictionary
             * mesh.prop(name, val) is equivalent to mesh.add_property(name, val) if val is not
               None
             * mesh.prop(name, None) is equivalent to mesh.remove_property(name)'''
        if arg is Ellipsis:
            if isinstance(name, dict):
                self.add_property(name)
            else:
                return self.property_value(name)
        else:
            self.add_property(name, arg)
    
    def map_vertices(self, f, merge=None):
        '''mesh.map_vertices(f) yields the result of mapping the function f over all vertices in
           the given mesh. For each vertex, f is called with a dictionary as the argument; the keys
           and values in the dictionary are the property names and property values for the
           vertices; additionally, the following properties are added to the dictionary:
             * \'vertex_label\': the label of the given vertex
             * \'coordinate\': the coordinate of the given vertex
           If map_vertices is called with the optional argument merge, a dictionary may be given;
           this dictionary must be a valid properties dictionary (see add_property) and the
           properties specified in the dictionary will be merged with the mesh's properties during
           the map operation.'''
        if merge is None: merge = {}
        extra = {'vertex_label':self.vertex_labels,
                 'coordinate': self.coordinates.transpose().tolist()}
        return map(
            f,
            [dict(zip(extra.keys() + self.properties.keys() + merge.keys(), vals)) \
                 for vals in zip(*(extra.values() + self.properties.values() + merge.values()))])

    def where(self, f):
        '''mesh.where(f) yields a boolean mask in which any vertex for which f(p) yields true
           is given a True value in the mask and all other vertives are given False values. The
           function f should operate on a dict p which is identical to that passed to the method
           mesh.map_vertices.'''
        return self.vertex_labels[np.array(self.map_vertices(f)) == True]

    def reproject(self, X):
        '''mesh.reproject(X) yields a 2D coordinate matrix Xp with the same number of points as the
           3D coordinate matrix X such that the points have been projected identically as the
           projection used to create this 2D mesh. If this is not a mesh created with the
           CorticalMesh projection function, an error will be raised.'''
        projection_params = self.option('projection_parameters')
        if projection_params is None:
            raise ValueError('the given mesh was not created by projection from another mesh')
        if 'inverse_function' not in projection_params:
            raise ValueError('the projection method used has not defined inverse')
        ffn = projection_params['forward_function']
        return ffn(X, projection_params)
    def unproject(self, X):
        '''mesh.unproject(X) yields a 3D coordinate matrix Xu with the same number of points as the
           2D coordinate matrix X such that the points have been projected back from the 2D map to
           the 3D sphere in the opposite fashion as the projection that was used to create this 2D
           mesh. If this is not a mesh created with the CorticalMesh projection function, an error
           will be raised. This function is the inverse of reproject.'''
        projection_params = self.option('projection_parameters')
        if projection_params is None:
            raise ValueError('the given mesh was not created by projection from another mesh')
        if 'inverse_function' not in projection_params:
            raise ValueError('the projection method used has not defined inverse')
        ifn = projection_params['inverse_function']
        return ifn(X, projection_params)

    def unaddress(self, data):
        '''mesh.unaddress(addr) yields a coordinate matrix of points on the given mesh that are
           located at the mesh addresses given in addr. The addr matrix should be generated from
           a mesh.address(points) method call; the two meshes must be topologically equivalent
           or an error may be raised. The resulting matrix will always be sized 2 or 3 by n unless
           the address is of a single point, in which case that point is returned.'''
        # addresses are dictionaries that contain three fields: 'face_id', 'angle_fraction', and 
        # 'distance_fraction'; these may be numbers or equally sized lists
        if self.coordinates.shape[0] == 2:
            # In this case, we unaddress on the sphere, then reproject...
            smesh = self.meta('source_mesh')
            if smesh is None: raise ValueError('2D mesh has no source mesh!')
            return self.reproject(smesh.unaddress(data))
        if not isinstance(data, dict):
            raise ValueError('address data must be a dictionary')
        if 'face_id' not in data: raise ValueError('address must contain face_id')
        if 'angle_fraction' not in data: raise ValueError('address must contain angle_fraction')
        if 'distance_fraction' not in data: raise ValueError('address must contain distance_fraction')
        face_id = data['face_id']
        angle_fraction = data['angle_fraction']
        distance_fraction = data['distance_fraction']
        if all(hasattr(x, '__iter__') for x in (face_id, angle_fraction, distance_fraction)):
            if len(face_id) != len(angle_fraction) or len(face_id) != len(distance_fraction):
                raise ValueError('invalid address specification (unmatched lengths)')
            return np.asarray(
                [geo.unaddress_triangle(self.face_coordinates[:,:,fid], [t, r])
                 for (fid,t,r) in zip(face_id, angle_fraction, distance_fraction)]
                ).T
        else:
            return geo.triangle_unaddress(self.face_coordinates[:, :, face_id],
                                          [angle_fraction, distance_fraction])
    def address(self, data):
        '''
        mesh.address(X) yields a dictionary containing the address or addresses of the point or
        points given in the vector or coordinate matrix X. Addresses specify a single unique 
        topological location on the mesh such that deformations of the mesh will address the same
        points differently. To convert a point from one mesh to another isomorphic mesh, you can
        address the point in the first mesh then unaddress it in the second mesh. Note that an
        address may only be obtained from a spherical or 2D mesh.
        '''
        if self.coordinates.shape[0] == 2:
            smesh = self.meta('source_mesh')
            if smesh is None: raise ValueError('2D mesh has no source mesh!')
            return smesh.address(self.unproject(data))
        else:
            reg = self.meta('registration')
            if reg is None: raise ValueError('mesh has no registration!')
            return reg.address(data)

    def option(self, opt):
        '''mesh.option(x) yields the value of the option f in the given mesh. If x is not an option
           then None is returned.  If x is a list, then the list is automatically mapped.'''
        opts = self.options
        if isinstance(opt, list):
            return map(lambda o: self.option(o), opt)
        elif opt in opts:
            return opts[opt]
        else:
            return None

    def meta(self, arg):
        '''mesh.meta(x) yields the meta-data element with name x in the given mesh. Note that meta
           is stored as an option in meshes.'''
        if isinstance(arg, list):
            return map(lambda x: self.meta(x), arg)
        else:
            return self.options.get('meta_data', {}).get(arg, None)


    ################################################################################################
    # Importers
    # These are static import methods
    
    @staticmethod
    def load(file):
        (coords, faces) = fsio.read_geometry(file)
        return CorticalMesh(coords, faces, source_file=file)


####################################################################################################
# cortex_to_ribbon converter and related functions

def cortex_to_mrvolume(mesh, property):
    '''cortex_to_mrvolume(mesh, property) yields a volume (represented as a numpy array) with the
         given property projected into the volume along the surface.
       cortex_to_mrvolume([mesh1, mesh2...], property) yields the volume in which all of the given
         meshes are projected into the volume simultaneously.

       The argument property may be either the name of a property attached to all the given meshes
       or a list or numpy array of values with the same length as the vertices in all the given 
       meshes.

       The following options are accepted:
         * method (default: 'weighted-mean') specifies how a voxel's contained vertices determine
           the value assigned to the voxel. This may be a callable object f(x0, X, Y) where the
           first argument x0 is the center of the voxel, the second argument X is a 3 x n numpy
           matrix of the (x, y, z) coordinates of each vertex in the voxel, and Y is a numpy vector
           of the property values associated with the vertices in X. Alternately, method may be
           set to any of the following:
            * 'weighted-mean' uses a Gaussian-weighted mean with the radius of a voxel as the 
              standard deviation of the Gaussian.
            * 'mean' uses mean(Y) as the value.
            * 'nearest' uses the stritly nearest vertex to the center of the voxel.
         * dimensions (default: [256, 256, 256]) specifies the size of the volume to be used. The
           size may be given in three ways: (1) as an list of widths [w, h, d], which are taken to
           be the number of voxels in each dimension with the voxels spanning from -[w h d]/2 to
           [w h d]/2, colloquially; (2) as a list of min/max pairs [[wmin wmax], ...] in which case
           the voxel width is assumed to be 1; or (3) as a list of min/max/n triples in which n is
           the number of voxels along that dimension.
         * origin (default: 'automatic') specifies where the (0,0,0) coordinate of the given cortex
           lines up in the volume coordinate system. If the default, 'automatic', is used, then the
           origin or the mean of the coordinate system (if the origin is not in the system) is
           used.
         * orientation (default: 'identity') specifies how the volume is oriented relative to the
           cortex. The default, 'identity', results in an identity matrix being used. Specifically,
           this argument should be a 3x3 matrix whose inverse is applied to the (x,y,z) coordinates
           of the cortical surface vertices before their containing voxels are determined. For a
           FreeSurfer surface, this matrix is equivalent to [R_row, A_row, S_row] where R_row,
           A_row, and S_row are unit vectors pointing in the direction of the right, anterior, and
           superior directions in the directions implied by array-indexing order.
         * default (default: 0) specifies what the default value in the volume should be.'''
    # First, process the args...

    ## Process the mesh arg; make sure the meshes are safe for the transform and make sure they are
    #  in a list
    if isinstance(mesh, CorticalMesh) and mesh.coordinates.shape[0] == 3:
        meshes = [mesh]
    elif hasattr(mesh, '__len__') and hasattr(mesh, '__iter__') \
            and len(mesh) > 0  \
            and all(isinstance(m, CorticalMesh) for m in mesh):
        if all(3 == m.coordinates.shape[0] for m in mesh):
            meshes = mesh
        else:
            raise ValueError('meshes must be a list of 3D cortical meshes')
    else:
        raise ValueError('meshes must be a 3D cortical mesh or a list of such meshes')
    ## Process the property arg; make sure the property exists in at least one of the cortices
    #  and filter the mesh list down to just those that do have it.
    #  We also go ahead and make the master vertices and properties matrices since we have all
    #  of that data now.
    if isinstance(property, basestring):
        # This is a property name; filter the mesh list to those that have it
        meshes = [m for m in meshes if property in m.properties]
        if len(meshes) == 0:
            raise ValueError('given property was not found in any provided mesh')
        properties = [m.prop(property) for m in meshes]
    elif hasattr(property, '__len__') and hasattr(property, '__iter__'):
        n = len(property)
        if not all([n == len(m.vertex_labels) for m in meshes]):
             raise ValueError('property lists must be the same length as all mesh vertex labels')
        properties = [property for m in meshes]
    else:
        raise ValueError('property list must be either a string or a list')

    #here
    return True



####################################################################################################
# neuropythy/geometry/__init__.py
# This file defines common rotation functions that are useful with cortical mesh spheres, such as
# those produced with FreeSurfer.

import numpy as np
import math

def normalize(u):
    '''
    normalize(u) yields a vetor with the same direction as u but unit length, or, if u has zero
    length, yields u.
    '''
    unorm = np.linalg.norm(u)
    if unorm == 0:
        return u
    return np.asarray(u) / unorm

def vector_angle_cos(u, v):
    '''
    vector_angle_cos(u, v) yields the cosine of the angle between the two vectors u and v. If u
    or v (or both) is a (d x n) matrix of n vectors, the result will be a length n vector of the
    cosines.
    '''
    return (u * v).sum(0) / np.sqrt((np.asarray(u) ** 2).sum(0) * (np.asarray(v) ** 2).sum(0))

def vector_angle(u, v, direction=None):
    '''
    vector_angle(u, v) yields the angle between the two vectors u and v. The optional argument 
    direction is by default None, which specifies that the smallest possible angle between the
    vectors be reported; if the vectors u and v are 2D vectors and direction parameters True and
    False specify the clockwise or counter-clockwise directions, respectively; if the vectors are
    3D vectors, then direction may be a 3D point that is not in the plane containing u, v, and the
    origin, and it specifies around which direction (u x v or v x u) the the counter-clockwise angle
    from u to v should be reported (the cross product vector that has a positive dot product with
    the direction argument is used as the rotation axis).
    '''
    if direction is None:
        return np.arccos(vector_angle_cos(u, v))
    elif direction is True:
        return np.arctan2(v[1], v[0]) - np.arctan2(u[1], u[0])
    elif direction is False:
        return np.arctan2(u[1], u[0]) - np.arctan2(v[1], v[0])
    else:
        axis1 = normalize(u)
        axis2 = normalize(np.cross(u, v))
        if np.dot(axis2, direction) < 0:
            axis2 = -axis2
        return np.arctan2(np.dot(axis2, v), np.dot(axis1, v))

def rotation_matrix_3D(u, th):
    """
    rotation_matrix_3D(u, t) yields a 3D numpy matrix that rotates any vector about the axis u
    t radians counter-clockwise.
    """
    # normalize the axis:
    u = normalize(u)
    # We use the Euler-Rodrigues formula;
    # see https://en.wikipedia.org/wiki/Euler-Rodrigues_formula
    a = math.cos(0.5 * th)
    s = math.sin(0.5 * th)
    (b, c, d) = -s * u
    (a2, b2, c2, d2) = (a*a, b*b, c*c, d*d)
    (bc, ad, ac, ab, bd, cd) = (b*c, a*d, a*c, a*b, b*d, c*d)
    return np.array([[a2 + b2 - c2 - d2, 2*(bc + ad),         2*(bd - ac)],
                     [2*(bc - ad),       a2 + c2 - b2 - d2,   2*(cd + ab)],
                     [2*(bd + ac),       2*(cd - ab),         a2 + d2 - b2 - c2]])

def rotation_matrix_2D(th):
    '''
    rotation_matrix_2D(th) yields a 2D numpy rotation matrix th radians counter-clockwise about the
    origin.
    '''
    s = np.sin(th)
    c = np.cos(th)
    return np.array([[c, -s], [s, c]])

# construct a rotation matrix of vector u to vector b around center
def alignment_matrix_3D(u, v):
    '''
    alignment_matrix_3D(u, v) yields a 3x3 numpy array that rotates the vector u to the vector v
    around the origin.
    '''
    # normalize both vectors:
    u = normalize(u)
    v = normalize(v)
    # get the cross product of u cross v
    w = np.cross(u, v)
    # the angle we need to rotate
    th = vector_angle(u, v)
    # we rotate around this vector by the angle between them
    return rotation_matrix_3D(w, th)

def alignment_matrix_2D(u, v):
    '''
    alignment_matrix_2D(u, v) yields a 2x2 numpy array that rotates the vector u to the vector v
    around the origin.
    '''
    return rotation_matrix_2D(vector_angle_2D(u, v, direction=True))

def line_intersection_2D(((x1,y1),(x2,y2)), ((x3,y3),(x4,y4))):
    '''
    line_intersection((a, b), (c, d)) yields the intersection point between the lines that pass
    through the given pairs of points.
    '''
    dx12 = (x1 - x2)
    dx34 = (x3 - x4)
    dy12 = (y1 - y2)
    dy34 = (y3 - y4)
    denom = dx12*dy34 - dy12*dx34
    q12 = (x1*y2 - y1*x2) / denom
    q34 = (x3*y4 - y3*x4) / denom
    return (q12*dx34 - q34*dx12, q12*dy34 - q34*dy12)

def triangle_area(a,b,c):
    '''
    triangle_area(a, b, c) yields the area of the triangle whose vertices are given by the points a,
    b, and c.
    '''
    if len(a) == 2:
        return 0.5*(a[0]*(b[1] - c[1]) + b[0]*(c[1] - a[1]) + c[0]*(a[1] - b[1]))
    else:
        mtx = alignment_matrix_3D(np.cross(np.asarray(b) - a, np.asarray(c) - a), [0,0,1])[0:1]
        return triangle_area(np.dot(mtx, a), np.dot(mtx, b), np.dot(mtx, c))

def triangle_address(fx, pt):
    '''
    triangle_address(FX, P) yields an address coordinate (t,r) for the point P in the triangle
    defined by the (3 x d)-sized coordinate matrix FX, in which each row of the matrix is the
    d-dimensional vector representing the respective triangle vertx for triangle [A,B,C]. The
    resulting coordinates (t,r) (0 <= t <= 1, 0 <= r <= 1) address the point P such that, if t gives
    the fraction of the angle from vector AB to vector AC that is made by the angle between vectors
    AB and AP, and r gives the fraction ||AP||/||AR|| where R is the point of intersection between
    lines AP and BC. If P is a (d x n)-sized matrix of points, then a (2 x n) matrix of addresses
    is returned.
    '''
    fx = np.asarray(fx)
    pt = np.asarray(pt)
    # The triangle vectors...
    ab = fx[1] - fx[0]
    ac = fx[2] - fx[0]
    bc = fx[2] - fx[1]
    ap = np.asarray([pt_i - a_i for (pt_i, a_i) in zip(pt, fx[0])])
    # get the unnormalized distance...
    r = np.sqrt((ap ** 2).sum(0))
    # now we can find the angle...
    unit = 1 - r.astype(bool)
    t0 = vector_angle(ab, ac)
    t = vector_angle(ap + [ab_i * unit for ab_i in ab], ab)
    sint = np.sin(t)
    sindt = np.sin(t0 - t)
    # finding r0 is tricker--we use this fancy formula based on the law of sines
    q0 = np.sqrt((bc ** 2).sum(0))          # B->C distance
    beta = vector_angle(-ab, bc)            # Angle at B
    sinGamma = np.sin(math.pi - beta - t0)
    sinBeta  = np.sin(beta)
    r0 = q0 * sinBeta * sinGamma / (sinBeta * sindt + sinGamma * sint)
    return np.asarray([t/t0, r/r0])

def triangle_unaddress(fx, tr):
    '''
    triangle_unaddress(FX, tr) yields the point P, inside the reference triangle given by the 
    (3 x d)-sized coordinate matrix FX, that is addressed by the address coordinate tr, which may
    either be a 2-d vector or a (2 x n)-sized matrix.
    '''
    fx = np.asarray(fx)
    tr = np.asarray(tr)
    # the triangle vectors...
    ab = fx[1] - fx[0]
    ac = fx[2] - fx[0]
    bc = fx[2] - fx[1]
    return np.asarray([ax + tr[1]*(abx + tr[0]*bcx) for (ax, bcx, abx) in zip(fx[0], bc, ab)])



####################################################################################################
# neuropythy/topology/__init__.py
# Tools for transferring data from one cortical surface to another via spherical registered 
# topologies.
# By Noah C. Benson

import numpy as np
import numpy.linalg
import scipy as sp
import scipy.spatial as space
import os, math
from pysistence import make_dict

from neuropythy.immutable import Immutable
import neuropythy.geometry as geo

class Topology:
    '''
    Topology(triangles, registrations) constructs a topology object object with the given triangle
    mesh, defined by a 3xn matrix of triangle indices, and with the registration coordinate matrices
    given in the dictionary registrations. This class should only be instantiated by the neuropythy
    library and should generally not be constructed directly. See Hemisphere.topology objects to
    access a subject's topologies.
    '''

    def __init__(self, triangles, registrations):
        triangles = np.asarray(triangles)
        self.triangles = triangles if triangles.shape[1] == 3 else triangles.T
        self.vertex_count = np.max(triangles.flatten())
        self.registrations = make_dict({name: Registration(self, np.asarray(coords))
                                        for (name, coords) in registrations.iteritems()
                                        if coords is not None})
    def __repr__(self):
        return 'Topology(<%d triangles>, <%d vertices>)' % (self.triangles.shape[0],
                                                            self.vertex_count)
    def register(self, name, coords):
        '''
        topology.register(name, coordinates) adds the given registration to the given topology.
        '''
        coords = np.asarray(coords)
        self.registrations = self.registrations.using(
            name,
            Registration(coords if coords.shape[0] == 3 else coords.T))
        return self
    def interpolate_from(self, topo, data, mask=None, null=None, method='automatic', n_jobs=1):
        '''
        topology.interpolate_from(topo, data) yields a numpy array of the data interpolated from
        the given array, data, which must contain the same number of elements as there are points in
        the topology object topo, to the coordinates in the given Topology object topology. Note
        that in order for an interpolation to occur, the two topologies, topology and topo, must
        have a shared registration; i.e., a registration with the same name.
        
        The following options are accepted:
          * mask (default: None) indicates that the given True/False or 0/1 valued list/array should
            be used; any point whose nearest neighbor (see below) is in the given mask will, instead
            of an interpolated value, be set to the null value (see null option).
          * null (default: None) indicates the value that should be placed in the returned result if
            either a vertex does not lie in any triangle or a vertex is masked out via the mask
            option.
          * smoothing (default: 2) assuming that the method is 'interpolate' or 'automatic', this
            is the exponent used to smooth the interpolated surface; 1 is pure linear interpolation
            while 2 represents a slightly smoother version of this. Note that this is not an order
            of interpolation option.
          * method (default: 'automatic') specifies what method to use for interpolation. The only
            currently supported methods are 'automatic' or 'nearest'. The 'nearest' method does not
            actually perform a nearest-neighbor interpolation but rather assigns to a destination
            vertex the value of the source vertex whose veronoi-like polygon contains the
            destination vertex; note that the term 'veronoi-like' is used here because it uses the
            Voronoi diagram that corresponds to the triangle mesh and not the true delaunay
            triangulation. The 'automatic' checks every destination vertex and assigns it the
            'nearest' value if that value would not be a number, otherwise it interpolates linearly
            within the vertex's source triangle.
          * n_jobs (default: 1) is passed along to the cKDTree.query method, so may be set to an
            integer to specify how many processors to use, or may be -1 to specify all processors.
        '''
        # find a shared registration:
        reg_name = next((k for k in topo.registrations.iterkeys() if k in self.registrations),
                        None)
        if reg_name is None:
            raise RuntimeError('Topologies do not share a matching registration!')
        return self.registrations[reg_name].interpolate_from(
            topo.registrations[reg_name], data,
            mask=mask, null=null, method=method, n_jobs=n_jobs);

class Registration(Immutable):
    '''
    A Registration object represents a configuration of a topology object. Registrations represent
    various coordinate configurations of a hemisphere; e.g., in FreeSurfer, any given hemisphere has
    at least the subject's native configuration, the fsaverage configuration, and the fsaverage_sym
    configuration. A registration may be used to sample data between subjects. Generally speaking,
    a registration object should only be constructed via other parts of the neuropythy library and
    not via Registration() directly.
    '''

    @staticmethod
    def __calculate_triangle_centers(triangles, coords):
        coords = np.asarray(coords)
        coords = coords if coords.shape[0] < coords.shape[1] else coords.T
        triangles = np.asarray(triangles)
        triangles = triangles if triangles.shape[0] < triangles.shape[1] else triangles.T
        return (coords[:,triangles[0]] + coords[:,triangles[1]] + coords[:,triangles[2]]).T / 3
        
    
    def __init__(self, topology, coordinates):
        coordinates = coordinates if coordinates.shape[0] < 4 else coordinates.T
        Immutable.__init__(
            self,
            {},
            {'topology':    topology,
             'coordinates': (coordinates / np.sqrt((coordinates ** 2).sum(0))).T},
            {'triangle_centers':
             (('topology','coordinates'),
              lambda topo,x: Registration.__calculate_triangle_centers(topo.triangles, x)),
             'triangle_hash': (('triangle_centers',), lambda x: space.cKDTree(x)),
             'vertex_hash':   (('coordinates',), lambda x: space.cKDTree(x))})
    def __repr__(self):
        return 'Registration(<%d triangles>, <%d vertices>)' % (self.topology.triangles.shape[0],
                                                                self.coordinates.shape[0])


    # True if the point is in the triangle, otherwise False; tri_no is an index into the triangle
    def _point_in_triangle(self, tri_no, pt):
        tri = self.coordinates[self.topology.triangles[tri_no]]
        return (np.dot(pt - tri[0], np.cross(tri[0], tri[1] - tri[0])) >= 0 and
                np.dot(pt - tri[1], np.cross(tri[1], tri[2] - tri[1])) >= 0 and
                np.dot(pt - tri[2], np.cross(tri[2], tri[0] - tri[2])) >= 0)

    def _find_triangle_search(self, x, k=24, searched=set([])):
        # This gets called when a container triangle isn't found; the idea is that k should
        # gradually increase until we find the container triangle; if k passes the max, then
        # we give up and assume no triangle is the container
        if k > 288: return None
        (d,near) = self.triangle_hash.query(x, k=k)
        near = [n for n in near if n not in searched]
        searched = searched.union(near)
        tri_no = next((k for k in near if self._point_in_triangle(k, x)), None)
        return (tri_no if tri_no is not None
                else self._find_triangle_search(x, k=(2*k), searched=searched))
        
    
    def nearest_vertex(self, pt):
        '''
        registration.nearest_vertex(pt) yields the id number of the nearest vertex in the given
        registration to the given point pt. If pt is an (n x dims) matrix of points, an id is given
        for each column of pt.
        '''
        (d,near) = self.vertex_hash.query(pt, k=1)
        return near
    def container(self, pt, k=2, n_jobs=1):
        '''
        registration.container(pt) yields the id number of the nearest triangle in the given
        registration to the given point pt. If pt is an (n x dims) matrix of points, an id is given
        for each column of pt.
        '''
        pt = np.asarray(pt)
        (d, near) = self.triangle_hash.query(pt, k=k) #n_jobs fails?
        if len(pt.shape) == 1:
            tri_no = next((k for k in near if self._point_in_triangle(k, pt)), None)
            return (tri_no if tri_no is not None
                    else self._find_triangle_search(pt, k=(2*k), searched=set(near)))
        else:
            return [(tri_no if tri_no is not None
                     else self._find_triangle_search(x, k=(2*k), searched=set(near_i)))
                    for (x, near_i) in zip(pt, near)
                    for tri_no in [next((k for k in near_i if self._point_in_triangle(k, x)),
                                        None)]]

    def interpolate_from(self, reg, data, 
                         smoothing=2, mask=None, null=None, method='automatic', n_jobs=1):
        '''
        registration.interpolate_from(reg, data) yields a numpy array of the data interpolated from
        the given array, data, which must contain the same number of elements as there are points in
        the Registration object reg, to the coordinates in the given Registration object,
        registration.
        
        The following options are accepted:
          * mask (default: None) indicates that the given True/False or 0/1 valued list/array should
            be used; any point whose nearest neighbor (see below) is in the given mask will, instead
            of an interpolated value, be set to the null value (see null option).
          * null (default: None) indicates the value that should be placed in the returned result if
            either a vertex does not lie in any triangle or a vertex is masked out via the mask
            option.
          * smoothing (default: 2) assuming that the method is 'interpolate' or 'automatic', this
            is the exponent used to smooth the interpolated surface; 1 is pure linear interpolation
            while 2 represents a slightly smoother version of this. Note that this is not an order
            of interpolation option.
          * method (default: 'automatic') specifies what method to use for interpolation. The only
            currently supported methods are 'automatic' or 'nearest'. The 'nearest' method does not
            actually perform a nearest-neighbor interpolation but rather assigns to a destination
            vertex the value of the source vertex whose veronoi-like polygon contains the
            destination vertex; note that the term 'veronoi-like' is used here because it uses the
            Voronoi diagram that corresponds to the triangle mesh and not the true delaunay
            triangulation. The 'automatic' checks every destination vertex and assigns it the
            'nearest' value if that value would not be a number, otherwise it interpolates linearly
            within the vertex's source triangle.
          * n_jobs (default: 1) is passed along to the cKDTree.query method, so may be set to an
            integer to specify how many processors to use, or may be -1 to specify all processors.
        '''
        # Okay, switch on method:
        if method == 'nearest':
            return self._interpolate_nearest(reg, data, mask, null, n_jobs)
        elif method == 'automatic':
            return self._interpolate_linear(reg, data, mask, null, smoothing, 12, n_jobs)

    # perform nearest-neighbor interpolation
    def _interpolate_nearest(self, reg, data, mask, null, n_jobs):
        # lookup the neighbors...
        (d, nei) = reg.vertex_hash.query(self.coordinates, k=1) #n_jobs fails? version problem?
        if mask is None:
            return [data[i] for i in nei]
        else:
            return [data[i] if mask[i] == 1 else null for i in nei]
    # perform linear interpolation
    def _interpolate_linear(self, reg, data, mask, null, smoothing, check_no, n_jobs):
        # first, find the triangle containing each point...
        tris = reg.topology.triangles
        data = np.asarray(data)
        ## we only query the nearest check_no triangles; otherwise we don't fine the container
        containers = reg.container(self.coordinates, k=check_no, n_jobs=n_jobs)
        # Okay, now we interpolate for each triangle
        if mask is None:
            return [(null if tri_no is None
                     else reg._interpolate_triangle(x, data, tris[tri_no], smoothing))
                    for (x, tri_no) in zip(self.coordinates, containers)]
        else:
            return [(null if tri_no is None or any(mask[u] == 0 for u in tris[tri_no])
                     else reg._interpolate_triangle(x, data, tris[tri_no], smoothing))
                    for (x, tri_no) in zip(self.coordinates, containers)]
    def _interpolate_triangle(self, x, data, tri_vertices, smoothing):
        # we'll want to project things down to 2 dimensions:
        mtx = geo.alignment_matrix_3D(x, [0,0,1])[0:2].T
        # Project out what we don't want
        corners = np.dot(self.coordinates[tri_vertices], mtx)
        x = np.asarray([0,0])
        # get the mini-triangles' areas
        a_area = geo.triangle_area(x, corners[1], corners[2]) ** smoothing
        b_area = geo.triangle_area(x, corners[2], corners[0]) ** smoothing
        c_area = geo.triangle_area(x, corners[0], corners[1]) ** smoothing
        # and do the interpolation:
        return np.dot([a_area, b_area, c_area], data[tri_vertices]) / (a_area + b_area + c_area)
        

    def address(self, data):
        '''
        reg.address(X) yields a dictionary containing the address or addresses of the point or
        points given in the vector or coordinate matrix X. Addresses specify a single unique 
        topological location on the mesh such that deformations of the mesh will address the same
        points differently. To convert a point from one mesh to another isomorphic mesh, you can
        address the point in the first mesh then unaddress it in the second mesh.
        '''
        # we have to have a topology and registration for this to work...
        data = np.asarray(data)
        if len(data.shape) == 1:
            face_id = self.container(data)
            (t, r) = geo.triangle_address(self.coordinates[self.topology.triangles[face_id]],
                                          data)
        else:
            data = data if data.shape[1] == 3 else data.T
            face_id = self.container(data)
            (t, r) = np.asarray([geo.triangle_address(self.coordinates[tri], x)
                                 for (tri,x) in zip(self.topology.triangles[face_id], data)]).T
        # And return the dictionary
        return {'face_id': face_id, 'angle_fraction': t, 'distance_fraction': r}
                



# __init__.py

'''Tools for analyzing and registering cortical meshes.'''

from cortex import cortex_to_mrvolume
from freesurfer import freesurfer_subject
from registration import mesh_register, register_retinotopy

# Version information...
_version_major = 0
_version_minor = 1
_version_micro = 0
__version__ = "%s.%s.%s" % (_version_major, _version_minor, _version_micro)

description = 'Integrate Python environment with FreeSurfer and perform mesh registration'
    



####################################################################################################
# neuropythy/freesurfer/__init__.py
# This file defines the FreeSurfer tools that are available as part of neuropythy.

from .subject import (Subject, Hemisphere, 
                      cortex_to_ribbon_map, cortex_to_ribbon, cortex_to_ribbon_map_lines)

# how to construct a freesurfer subject:
def freesurfer_subject(name, subjects_dir=None):
    '''
    freesurfer_subject(name) yields a freesurfer Subject object for the subject with the given name.
    If the given subject cannot be found, then a ValueError is raised. The option subjects_dir may
    be given to specify the location of the subject directory.
    '''
    if subjects_dir is None:
        return Subject(name)
    else:
        return Subject(name, subjects_dir=subjects_dir)



####################################################################################################
# registration.py
# Tools for registering the cortical surface to a particular potential function
# By Noah C. Benson

import numpy as np
import scipy as sp
import os, sys
from math import pi
from numbers import Number
from neuropythy.cortex import CorticalMesh
from neuropythy.freesurfer import freesurfer_subject
from neuropythy.topology import Registration
from pysistence import make_dict
from array import array

from py4j.java_gateway import (launch_gateway, JavaGateway, GatewayParameters)

# Java start:
_java_port = None
_java = None
def init_registration():
    global _java, _java_port
    if _java is not None: return
    _java_port = launch_gateway(
        classpath=os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))),
            'lib', 'nben', 'target', 'nben-standalone.jar'),
        javaopts=['-Xmx2g'],
        die_on_exit=True)
    _java = JavaGateway(gateway_parameters=GatewayParameters(port=_java_port))

# These are dictionaries of all the details we have about each of the possible arguments to the
# mesh_register's field argument:
_parse_field_data_types = {
    'mesh': ['newStandardMeshPotential', ['edge_scale', 1.0], ['angle_scale', 1.0], 'F', 'X'],
    'edge': {
        'harmonic':      ['newHarmonicEdgePotential',   ['scale', 1.0], ['order', 2.0], 'F', 'X'],
        'lennard-jones': ['newLJEdgePotential',         ['scale', 1.0], ['order', 2.0], 'F', 'X']},
    'angle': {
        'harmonic':      ['newHarmonicAnglePotential',  ['scale', 1.0], ['order', 2.0], 'F', 'X'],
        'lennard-hones': ['newLJAnglePotential',        ['scale', 1.0], ['order', 2.0], 'F', 'X'],
        'infinite-well': ['newWellAnglePotential',      ['scale', 1.0], ['order', 2.0], 
                                                        ['min',   0.0], ['max',   pi],  'F', 'X']},
    'anchor': {
        'harmonic':      ['newHarmonicAnchorPotential', ['scale', 1.0], ['shape', 2.0], 0, 1, 'X'],
        'gaussian':      ['newGaussianAnchorPotential', ['scale', 1.0], ['shape', 2.0], 
                                                        ['sigma', 1.0], 0, 1, 'X']},
    'perimeter': {
        'harmonic':   ['newHarmonicPerimeterPotential', ['scale', 1.0], ['shape', 2.0], 'F', 'X']}};

def serialize_numpy(m, t):
    '''
    serialize_numpy(m, type) converts the numpy array m into a byte stream that can be read by the
    nben.util.Py4j Java class. The function assumes that the type of the array needn't be encoded
    in the bytearray itself. The bytearray will begin with an integer, the number of dimensions,
    followed by that number of integers (the dimension sizes themselves) then the bytes of the
    array, flattened.
    The argument type gives the type of the array to be transferred and must be 'i' for integer or
    'd' for double (or any other string accepted by array.array()).
    '''
    # Start with the header: <number of dimensions> <dim1-size> <dim2-size> ...
    header = array('i', [len(m.shape)] + list(m.shape))
    # Now, we can do the array itself, just flattened
    body = array(t, m.flatten().tolist())
    # Wrap bytes if necessary...
    if sys.byteorder != 'big':
        header.byteswap()
        body.byteswap()
    # And return the result:
    return bytearray(header.tostring() + body.tostring())

def to_java_doubles(m):
    '''
    to_java_doubles(m) yields a java array object for the vector or matrix m.
    '''
    global _java
    if _java is None: init_registration()
    m = np.asarray(m)
    dims = len(m.shape)
    if dims > 2: raise ValueError('1D and 2D arrays supported only')
    bindat = serialize_numpy(m, 'd')
    return (_java.jvm.nben.util.Numpy.double2FromBytes(bindat) if dims == 2
            else _java.jvm.nben.util.Numpy.double1FromBytes(bindat))
def to_java_ints(m):
    '''
    to_java_ints(m) yields a java array object for the vector or matrix m.
    '''
    global _java
    if _java is None: init_registration()
    m = np.asarray(m)
    dims = len(m.shape)
    if dims > 2: raise ValueError('1D and 2D arrays supported only')
    bindat = serialize_numpy(m, 'i')
    return (_java.jvm.nben.util.Numpy.int2FromBytes(bindat) if dims == 2
            else _java.jvm.nben.util.Numpy.int1FromBytes(bindat))
def to_java_array(m):
    '''
    to_java_array(m) yields to_java_ints(m) if m is an array of integers and to_java_doubles(m) if
    m is anything else. The numpy array m is tested via numpy.issubdtype(m.dtype, numpy.int64).
    '''
    m = np.asarray(m)
    return to_java_ints(m) if np.issubdtype(m.dtype, np.int64) else to_java_doubles(m)
        
def _parse_field_function_argument(argdat, args, faces, coords):
    # first, see if this is an easy one...
    if argdat == 'F':
        return faces
    elif argdat == 'X':
        return coords
    elif isinstance(argdat, (int, long)):
        return to_java_array(args[argdat])
    # okay, none of those; must be a list with a default arg
    argname = argdat[0]
    argdflt = argdat[1]
    # see if we can find such an arg...
    for i in range(len(args)):
        if isinstance(args[i], basestring) and args[i].lower() == argname.lower():
            return args[i+1] if isinstance(args[i+1], Number) else to_java_array(args[i+1])
    # did not find the arg; use the default:
    return argdflt

def _parse_field_argument(instruct, faces, coords):
    global _java
    if _java is None: init_registration()
    if isinstance(instruct, basestring):
        insttype = instruct
        instargs = []
    elif type(instruct) in [list, tuple]:
        insttype = instruct[0]
        instargs = instruct[1:]
    else:
        raise RuntimeError('potential field instruction must be list/tuple or string')
    # look this type up in the types data:
    insttype = insttype.lower()
    if insttype not in _parse_field_data_types:
        raise RuntimeError('Unrecognized field data type: ' + insttype)
    instdata = _parse_field_data_types[insttype]
    # if the data is a dictionary, we must parse on the next arg
    if isinstance(instdata, dict):
        shape_name = instargs[0].lower()
        instargs = instargs[1:]
        if shape_name not in instdata:
            raise RuntimeError('Shape ' + shape_name + ' not supported for type ' + insttype)
        instdata = instdata[shape_name]
    # okay, we have a list of instructions... find the java method we are going to call...
    java_method = getattr(_java.jvm.nben.mesh.registration.Fields, instdata[0])
    # and parse the arguments into a list...
    java_args = [_parse_field_function_argument(a, instargs, faces, coords) for a in instdata[1:]]
    # and call the function...
    return java_method(*java_args)

# parse a field potential argument and return a java object that represents it
def _parse_field_arguments(arg, faces, coords):
    '''See mesh_register.'''
    if not isinstance(arg, list):
        raise RuntimeError('field argument must be a list of instructions')
    pot = [_parse_field_argument(instruct, faces, coords) for instruct in arg]
    # make a new Potential sum unless the length is 1
    if len(pot) <= 1:
        return pot[0]
    else:
        sp = _java.jvm.nben.mesh.registration.Fields.newSum()
        for field in pot: sp.addField(field)
        return sp

# The mesh_register function
def mesh_register(mesh, field, max_steps=25000, max_step_size=0.1, max_pe_change=1):
    '''
    mesh_register(mesh, field) yields the mesh that results from registering the given mesh by
    minimizing the given potential field description over the position of the vertices in the
    mesh. The mesh argument must be a CorticalMesh (see neuropythy.cortex) such as can be read
    from FreeSurfer using the neuropythy.freesurfer.Subject class. The field argument must be
    a list of field names and arguments; with the exception of 'mesh' (or 'standard'), the 
    arguments must be a list, the first element of which is the field type name, the second
    element of which is the field shape name, and the final element of which is a dictionary of
    arguments accepted by the field shape.

    The following are valid field type names:
      * 'mesh' : the standard mesh potential, which includes an edge potential, an angle
        potential, and a perimeter potential. Accepts no arguments, and must be passed as a
        single string instead of a list.
      * 'edge': an edge potential field in which the potential is a function of the change in the
        edge length, summed over each edge in the mesh.
      * 'angle': an angle potential field in which the potential is a function of the change in
        the angle measure, summed over all angles in the mesh.
      * 'perimeter': a potential that depends on the vertices on the perimeter of a 2D mesh
        remaining in place; the potential changes as a function of the distance of each perimeter
        vertex from its reference position.
      * 'anchor': a potential that depends on the distance of a set of vertices from fixed points
        in space. After the shape name second argument, an anchor must be followed by a list of
        vertex ids then a list of fixed points to which the vertex ids are anchored:
        ['anchor', shape_name, vertex_ids, fixed_points, args...].

    The following are valid shape names:
      * 'harmonic': a harmonic function with the form (c/q) * abs(x - x0)^q.
        Parameters: 
          * 'scale', the scale parameter c; default: 1.
          * 'order', the order parameter q; default: 2.
      * 'Lennard-Jones': a Lennard-Jones function with the form c (1 + (r0/r)^q - 2(r0/r)^(q/2));
        Parameters:
          * 'scale': the scale parameter c; default: 1. 
          * 'order': the order parameter q; default: 2.
      * 'Gaussian': A Gaussian function with the form c (1 - exp(-0.5 abs((x - x0)/s)^q))
        Parameters:
          * 'scale': the scale parameter c; default: 1.
          * 'order': the order parameter q; default: 2.
          * 'sigma': the standard deviation parameter s; default: 1.
      * 'infinite-well': an infinite well function with the form 
        c ( (((x0 - m)/(x - m))^q - 1)^2 + (((M - x0)/(M - x))^q - 1)^2 )
        Parameters:
          * 'scale': the scale parameter c; default: 1.
          * 'order': the order parameter q; default: 0.5.
          * 'min': the minimum value m; default: 0.
          * 'max': the maximum value M; default: pi.

    Options: The following optional arguments are accepted.
      * max_steps (default: 25000) the maximum number of steps to minimize for.
      * max_step_size (default: 0.1) the maximum distance to allow a vertex to move in a single
        minimization step.
      * max_pe_change: the maximum fraction of the initial potential value that the minimizer
        should minimize away before returning; i.e., 0 indicates that no minimization should be
        allowed while 0.9 would indicate that the minimizer should minimize until the potential
        is 10% or less of the initial potential.

    Examples:
      registered_mesh = mesh_register(
         mesh,
         [['edge', 'harmonic', 'scale', 0.5], # slightly weak edge potential
          ['angle', 'infinite-well'], # default arguments for an infinite-well angle potential
          ['anchor', 'Gaussian', [1, 10, 50], [[0.0, 0.0], [1.1, 1.1], [2.2, 2.2]]]],
         max_step_size=0.05,
         max_steps=10000)
    '''
    global _java
    if _java is None: init_registration()
    # Sanity checking.
    # First, make sure that the arguments are all okay:
    if not isinstance(mesh, CorticalMesh):
        raise RuntimeError('mesh argument must be an instance of neuropythy.cortex.CorticalMesh')
    if not isinstance(max_steps, (int, long)) or max_steps < 1:
        raise RuntimeError('max_steps argument must be a positive integer')
    if not isinstance(max_steps, (float, int, long)) or max_step_size <= 0:
        raise RuntimeError('max_step_size must be a positive number')
    if not isinstance(max_pe_change, (float, int, long)) or max_pe_change <= 0 or max_pe_change > 1:
        raise RuntimeError('max_pe_change must be a number x such that 0 < x <= 1')
    # Parse the field argument.
    faces  = to_java_ints([mesh.index[frow] for frow in mesh.faces])
    coords = to_java_doubles(mesh.coordinates)
    potential = _parse_field_arguments(field, faces, coords)
    # Okay, that's basically all we need to do the minimization...
    minimizer = _java.jvm.nben.mesh.registration.Minimizer(potential, coords)
    minimizer.step(float(max_pe_change), int(max_steps), float(max_step_size))
    return np.array(minimizer.getX())

# How we construct a Schira Model:
class SchiraModel:
    '''
    The SchiraModel class is a Python wrapper around the Java nben.neuroscience.SchiraModel class;
    it handles conversion from visual field angle to cortical surface coordinates and vice versa for
    the Banded Double-Sech model proposed in the following paper:
    Schira MM, Tyler CW, Spehar B, Breakspear M (2010) Modeling Magnification and Anisotropy in the
    Primate Foveal Confluence. PLOS Comput Biol 6(1):e1000651. doi:10.1371/journal.pcbi.1000651.
    '''

    # These are the accepted arguments to the model:
    default_parameters = {
        'A': 0.5,
        'B': 135.0,
        'lambda': 1.0,
        'psi': 0.15,
        'scale': [7.0, 8.0],
        'shear': [[1.0, -0.2], [0.0, 1.0]],
        'center': [-7.0, -2.0],
        'v1size': 1.2,
        'v2size': 0.6,
        'v3size': 0.4,
        'hv4size': 0.9,
        'v3asize': 0.9}
    # This function checks the given arguments to see if they are okay:
    def __check_parameters(self, parameters):
        # we don't care if there are extra parameters; we just make sure the given parameters make
        # sense, then return the full set of parameters
        opts = {
            k: parameters[k] if k in parameters else v
            for (k,v) in SchiraModel.default_parameters.iteritems()}
        return opts

    # This class is immutable: don't change the params to change the model; 
    # don't change the java object!
    def __setattr__(self, name, val):
        raise ValueError('The SchiraModel class is immutable; its objects cannot be edited')

    def __init__(self, **opts):
        global _java
        if _java is None:
            init_registration()
        # start by getting the proper parameters
        params = self.__check_parameters(opts)
        # Now, do the translations that we need...
        if isinstance(params['scale'], Number):
            params['scale'] = [params['scale'], params['scale']]
        if isinstance(params['shear'], Number) and params['shear'] == 0:
            params['shear'] = [[1, 0], [0, 1]]
        elif params['shear'][0][0] != 1 or params['shear'][1][1] != 1:
            raise RuntimeError('shear matrix [0,0] elements and [1,1] elements must be 1!')
        if isinstance(params['center'], Number) and params['center'] == 0:
            params['center'] = [0.0, 0.0]
        self.__dict__['parameters'] = make_dict(params)
        # Okay, let's construct the object...
        self.__dict__['_java_object'] = _java.jvm.nben.neuroscience.SchiraModel(
            params['A'],
            params['B'],
            params['lambda'],
            params['psi'],
            params['v1size'],
            params['v2size'],
            params['v3size'],
            params['hv4size'],
            params['v3asize'],
            params['center'][0],
            params['center'][1],
            params['scale'][0],
            params['scale'][1],
            params['shear'][0][1],
            params['shear'][1][0])
    
    def angle_to_cortex(self, theta, rho):
        global _java
        if _java is None: init_registration()
        iterTheta = hasattr(theta, '__iter__')
        iterRho = hasattr(rho, '__iter__')
        if iterTheta and iterRho:
            if len(theta) != len(rho):
                raise RuntimeError('Arguments theta and rho must be the same length!')
            return self._java_object.angleToCortex(to_java_doubles(theta), to_java_doubles(rho))
        elif iterTheta:
            return self._java_object.angleToCortex(to_java_doubles(theta),
                                                   to_java_doubles([rho for t in theta]))
        elif iterRho:
            return self._java_object.angleToCortex(to_java_doubles([theta for r in rho]),
                                                   to_java_doubles(rho))
        else:
            return self._java_object.angleToCortex(theta, rho)
    def cortex_to_angle(self, x, y):
        global _java
        iterX = hasattr(x, '__iter__')
        iterY = hasattr(y, '__iter__')
        if iterX and iterY:
            if len(x) != len(y):
                raise RuntimeError('Arguments x and y must be the same length!')
            return self._java_object.cortexToAngle(to_java_doubles(x), to_java_doubles(y))
        elif iterX:
            return self._java_object.cortexToAngle(to_java_doubles(x),
                                                   to_java_doubles([y for i in x]))
        elif iterY:
            return self._java_object.cortexToAngle(to_java_doubles([x for i in y]),
                                                   to_java_doubles(y))
        else:
            return self._java_object.cortexToAngle(x, y)

def schira_anchors(mesh, mdl,
                   polar_angle=None, eccentricity=None,
                   weight=None, weight_cutoff=None,
                   shape='Gaussian', suffix=None):
    '''
    schira_anchors(mesh, model) is intended for use with the mesh_register function and the 
    SchiraModel class; it yields a description of the anchor points that tie relevant vertices of
    the given mesh to points predicted by the given SchiraModel object, model.

    Options:
      * polar_angle (default None) specifies that the given data should be used in place of the
        'polar_angle' or 'PRF_polar_angle'  property values. The given argument must be numeric and
        the same length as the the number of vertices in the mesh. If None is given, then the
        property value of the mesh is used; if a list is given and any element is None, then the
        weight for that vertex is treated as a zero. If the option is a string, then the property
        value with the same name isused as the polar_angle data.
      * eccentricity (default None) specifies that the given data should be used in places of the
        'eccentricity' or 'PRF_eccentricity' property values. The eccentricity option is handled 
        virtually identically to the polar_angle option.
      * weight (default None) specifies that the weight or scale of the data; this is handled
        generally like the polar_angle and eccentricity options, but may also be 1, indicating that
        all vertices with polar_angle and eccentricity values defined will be given a weight of 1.
        If weight is left as None, then the function will check for 'weight',
        'variance_explained', 'PRF_variance_explained', and 'retinotopy_weight' values and will use
        the first found (in that order). If none of these is found, then a value of 1 is assumed.
      * weight_cutoff (default 0) specifies that the weight must be higher than the given value inn
        order to be included in the fit; vertices with weights below this value have their weights
        truncated to 0.
      * shape (default 'Gaussian') specifies the shape of the potential function (see mesh_register)
      * suffix (default None) specifies any additional arguments that should be appended to the 
        potential function description list that is produced by this function; i.e., schira_anchors
        produces a list, and the contents of suffix, if given and not None, are appended to that
        list (see mesh_register).

    Example:
    The schira_anchors function is intended for use with mesh_register, as follows:
    # Define our Schira Model:
    model = neuropythy.registration.SchiraModel()
    # Make sure our mesh has polar angle, eccentricity, and weight data:
    mesh.prop('polar_angle',  polar_angle_vertex_data);
    mesh.prop('eccentricity', eccentricity_vertex_data);
    mesh.prop('weight',       variance_explained_vertex_data);
    # register the mesh using the retinotopy and model:
    registered_mesh = neuropythy.registration.mesh_register(
       mesh,
       ['mesh', schira_anchors(mesh, model, weight_cutoff=0.2)],
       max_step_size=0.05,
       max_steps=10000)
    '''
    if not isinstance(mdl, SchiraModel):
        raise RuntimeError('given model is not a SchiraModel instance!')
    if not isinstance(mesh, CorticalMesh):
        raise RuntimeError('given mesh is not a CorticalMesh object!')
    n = len(mesh.vertex_labels)
    # make sure we have our polar angle/eccen/weight values:
    polar_angle = polar_angle if polar_angle is not None \
        else mesh.prop('polar_angle') if mesh.has_property('polar_angle') \
        else mesh.prop('PRF_polar_angle') if mesh.has_property('PRF_polar_angle') \
        else None
    if polar_angle is None:
        raise RuntimeError('No polar angle data given to schira_anchors!')
    if isinstance(polar_angle, dict):
        # a dictionary is okay, we just need to fix it to a list:
        tmp = polar_angle
        polar_angle = [tmp[i] if i in tmp else None for i in range(n)]
    if len(polar_angle) != n:
        raise RuntimeError('Polar angle data has incorrect length!')
    # Now Polar Angle...
    eccentricity = eccentricity if eccentricity is not None \
        else mesh.prop('eccentricity') if mesh.has_property('eccentricity') \
        else mesh.prop('PRF_eccentricity') if mesh.has_property('PRF_eccentricity') \
        else None
    if eccentricity is None:
        raise RuntimeError('No eccentricity data given to schira_anchors!')
    if isinstance(eccentricity, dict):
        tmp = eccentricity
        eccentricity = [tmp[i] if i in tmp else None for i in range(n)]
    if len(eccentricity) != n:
        raise RuntimeError('Eccentricity data has incorrect length!')
    # Now Weight...
    weight = weight if weight is not None \
        else mesh.prop('weight') if mesh.has_property('weight') \
        else mesh.prop('variance_explained') if mesh.has_property('variance_explained') \
        else mesh.prop('PRF_variance_explained') if mesh.has_property('PRF_variance_explained') \
        else mesh.prop('retinotopy_weight') if mesh.has_property('retinotopy_weight') \
        else None
    if weight is None:
        weight = 1
    if isinstance(weight, dict):
        tmp = weight
        weight = [tmp[i] if i in tmp else None for i in range(n)]
    if isinstance(weight, Number):
        weight = [weight for i in range(n)]
    if len(weight) != n:
        raise RuntimeError('Weight data has incorrect length!')
    # let's go through and fix up the weights/polar angles/eccentricities into appropriate lists
    if weight_cutoff is None:
        data = [[i, mdl.angle_to_cortex(polar_angle[i], eccentricity[i]), weight[i]]
                for i in range(n)
                if (polar_angle[i] is not None and eccentricity[i] is not None 
                    and weight[i] is not None and weight[i] != 0)]
    else:
        data = [[i, mdl.angle_to_cortex(polar_angle[i], eccentricity[i]), weight[i]]
                for i in range(n)
                if (polar_angle[i] is not None and eccentricity[i] is not None 
                    and weight[i] is not None and weight[i] >= weight_cutoff)]
    # okay, we've partially parsed the data that was given; now we can construct the final list of
    # instructions:
    return ['anchor', shape,
            [d[0] for d in data for k in range(len(d[1]))],
            np.asarray([pt for d in data for pt in d[1]]).T,
            'scale', [d[2] for d in data for k in range(len(d[1]))]
           ] + ([] if suffix is None else suffix)

_empirical_retinotopy_names = {
    'polar_angle':  ['PRF_polar_angle',  'empirical_polar_angle',  'measured_polar_angle'
                     'polar_angle'],
    'eccentricity': ['PRF_eccentricity', 'empirical_eccentricity', 'measured_eccentricity'
                     'eccentricity'],
    'weight':       ['PRF_variance_explained',       'PRF_weight',
                     'measured_variance_explained',  'measured_weight',
                     'empirical_variance_explained', 'empirical_weight'
                     'variance_explained',           'weight']}
# handy function for picking out properties automatically...
def empirical_retinotopy_data(hemi, retino_type):
    '''
    empirical_retinotopy_data(hemi, t) yields a numpy array of data for the given hemisphere object
    and retinotopy type t; it does this by looking at the properties in hemi and picking out any
    combination that is commonly used to denote empirical retinotopy data. These common names are
    stored in _empirical_retintopy_names, in order of preference, which may be modified.
    The argument t should be one of 'polar_angle', 'eccentricity', 'weight'.
    '''
    dat = _empirical_retinotopy_names[retino_type.lower()]
    return next((hemi.prop(s) for s in dat if hemi.has_property(s)), None)
    
def register_retinotopy(hemi,
                        schira_model=SchiraModel(), radius=pi/4.0,
                        polar_angle=None, eccentricity=None, weight=None, weight_cutoff=None,
                        sigma=0.5, edge_scale=1.0, angle_scale=1.0, schira_scale=15.0,
                        max_steps=30000, max_step_size=0.05,
                        registration_name='retinotopy'):
    '''
    register_retinotopy(hemi) yields the result of registering the given hemisphere's polar angle
    and eccentricity data to the SchiraModel, a registration in which the vertices are aligned with
    the Schira model of retinotopy. The registration is added to the hemisphere's topology unless
    the option registration_name is set to None.

    Options:
      * schira_model specifies the instance of the Schira model to use (default: SchiraModel()).
      * polar_angle, eccentricity, and weight specify the property names for the respective
        quantities; these may alternately be lists or numpy arrays of values. If weight is not given
        or found, then unity weight for all vertices is assumed. By default, each will check the
        hemisphere's properties for properties with compatible names; it will prefer the properties
        PRF_polar_angle, PRF_ecentricity, and PRF_variance_explained if possible.
      * weight_cutoff specifies the minimum value a vertex must have in the weight property in order
        to be considered as retinotopically relevant.
      * sigma specifies the standard deviation of the Gaussian shape for the Schira model anchors.
      * edge_scale, angle_scale, and schira_scale all specify the relative strengths of the various
        components of the potential field.
      * max_steps (default 30,000) specifies the maximum number of registration steps to run.
      * max_step_size (default 0.05) specifies the maxmim distance a single vertex is allowed to
        move in a single step of the minimization.
      * registration_name (default: 'retinotopy') specifies the name of the registration to register
        with the hemisphere's topology object.
    '''
    # Step 1: figure out what properties we're using...
    (ang, ecc, wgt) = [
        (hemi.prop(arg) if isinstance(arg, basestring) else
         arg            if hasattr(arg, '__iter__')    else
         None           if arg is not None             else
         empirical_retinotopy_data(hemi, argstr))
        for (arg, argstr) in [(polar_angle, 'polar_angle'),
                              (eccentricity, 'eccentricity'),
                              (weight, 'weight')]]
    if ang is None: raise ValueError('polar angle data not found')
    if ecc is None: raise ValueError('eccentricity data not found')
    ## we also want to make sure weight is 0 where there are none values
    wgt = np.asarray(
        [(0 if w is None else w)*(0 if a is None else 1)*(0 if e is None else 1)
         for (a,e,w) in zip(ang, ecc, [1 for e in ecc] if wgt is None else wgt)])
    ang = np.asarray([0 if a is None else a for a in ang])
    ecc = np.asarray([0 if e is None else e for e in ecc])
    # Step 2: get the properties over to the fsaverage_sym hemisphere
    lhemi = hemi if hemi.chirality == 'LH' else hemi.subject.RHX
    sym = freesurfer_subject('fsaverage_sym').LH
    sym.interpolate(lhemi, ang, apply='polar_angle')
    sym.interpolate(lhemi, ecc, apply='eccentricity')
    sym.interpolate(lhemi, wgt, apply='weight')
    # Step 3: make the projection
    msym = sym.projection(radius=radius)
    # Step 4: run the mesh registration
    r = mesh_register(
        msym,
        [['edge', 'harmonic', 'scale', edge_scale],
         ['angle', 'infinite-well', 'scale', angle_scale],
         ['perimeter', 'harmonic'],
         schira_anchors(msym, schira_model,
                        weight_cutoff=weight_cutoff,
                        suffix=['sigma', sigma, 'scale', schira_scale])],
        max_steps=max_steps,
        max_step_size=max_step_size)
    # Step 5: prepare the original subject's map for being warped over to the registration
    msub = lhemi.projection(mesh='fsaverage_sym', radius=radius)
    subvtcs = msub.vertex_labels
    msub.prop('polar_angle',  ang[subvtcs])
    msub.prop('eccentricity', ecc[subvtcs])
    msub.prop('weight',       wgt[subvtcs])
    ## okay, r is the registered coordinates; we can unproject them
    symrsphere = msym.unproject(r)
    ## make a new coordinate matrix for the registration
    symregcoords = np.array(sym.coordsinates, copy=True)
    symregcoords[:, msym.vertex_labels] = symrsphere
    symregmesh = sym.LH.surface(symregcoords)
    ## now, address the subject's coordinates in the original topology and unaddress in this
    ## new registered mesh
    addr = sym.topology.registrations['fsaverage_sym'].address(lhemi.coordinates[:,subvtcs])
    subrsphere = symregmesh.unaddress(addr)
    msub.coordinates = msub.reproject(subrsphere)
    if registration_name is not None:
        subregcoords = np.array(lhemi.coordinates, copy=True)
        subregcoords[:, subvtcs] = subrsphere
        hemi.topology.register(registration_name, subregcoords)
    # We return the subject's map
    return msub

# The topology and registration stuff is below:
class JavaTopology:
    '''
    JavaTopology(triangles, registrations) creates a topology object object with the given triangle
    mesh, defined by a 3xn matrix of triangle indices, and with the registration coordinate matrices
    given in the dictionary registrations. This class should only be instantiated by the neuropythy
    library and should generally not be constructed directly. See Hemisphere.topology objects to
    access a subject's topologies.
    '''
    def __init__(self, triangles, registrations):
        if _java is None:
            init_registration()
        # First: make a java object for the topology:
        faces = serialize_numpy(triangles.T, 'i')
        topo = _java.jvm.nben.geometry.spherical.MeshTopology.fromBytes(faces)
        # Okay, make our registration dictionary
        d = {k: topo.registerBytes(serialize_numpy(v, 'd'))
             for (k,v) in registrations.iteritems()}
        # That's all really
        self.__dict__['_java_object'] = topo
        self.__dict__['registrations'] = d
    def __getitem__(self, attribute):
        return self.registrations[attribute]
    def __setitem__(self, attribute, dat):
        self.registrations[attribute] = self._java_object.registerBytes(serialize_numpy(dat, 'd'))
    def keys(self):
        return self.registrations.keys()
    def iterkeys(self):
        return self.registrations.iterkeys()
    def values(self):
        return self.registrations.values()
    def itervalues(self):
        return self.registrations.itervalues()
    def items(self):
        return self.registrations.items()
    def iteritems(self):
        return self.registrations.iteritems()
    def __len__(self):
        return len(self.registrations)
    
    # These let us interpolate...
    def interpolate(fromtopo, data, order=2, fill=None):
        usable_keys = []
        for k in registrations.iterkeys():
            if k in fromtopo.registrations:
                usable_keys.append(k)
        if not usable_keys:
            raise RuntimeError('no registration found that links topologies')
        the_key = usable_keys[0]
        # Prep the data into java arrays
        jmask = serialize_numpy(np.asarray([1 if d is not None else 0 for d in data]), 'd')
        jdata = serialize_numpy(np.asarray([d if d is not None else 0 for d in data]), 'd')
        # okay, next step is to call out to the java...
        maskres = self._java_object.interpolateBytes(
            fromtopo.registrations[the_key],
            self.registrations[the_key].coordinates,
            order, jdata)
        datares = self._java_object.interpolateBytes(
            fromtopo.registrations[the_key],
            self.registrations[the_key].coordinates,
            order, jmask)
        # then interpret the results...
        return [datares[i] if maskres[i] == 1 else fill for i in range(len(maskres))]



####################################################################################################
# neuropythy/immutable/__init__.py
# Simple class to inherit for immutable lazily-loading classes.
# By Noah C. Benson

import itertools
import collections
from pysistence import make_dict
import pysistence

class Immutable:
    '''
    The Immutable class can be overloaded by any class that wishes to be an immutable lazily-loading
    class---Immutable facilitates that. It does so by overloading the standard dictionary functions
    and accepting a list of lazy loading instructions in its constructor. The constructor for the
    Immutable class expects three arguments plus any number of optional key-val assignments. The 
    optional key-val assignments are placed into the settable member 'options'. The first three 
    arguments are: Immutable(settable_vals, const_vals, lazy_vals).
      * settable_vals must be a dict of settable value names (strings) mapped to a function which
        takes as an argument the self object and the value being assigned and returns the value
        that should actually be assigned.
      * const_vals must be a dict of constant value names (strings) mapped to constant values; these
        will be constant member values of the object.
      * lazy_vals must be a dict of lazy value names (strings) mapped to a 2-tuple of (1) a tuple of
        argument names and (2) a function that accepts these arguments and yields the lazy value;
        the argument names may be any settable, const, or lazy value name.
    '''

    # Calculate lazy dependencies between values for this object
    def __lazy_deps(self):
        # we build up this dependency tree:
        dep_tree = dict()
        for (sname, sdata) in self._settable_vals.iteritems():
            deps = {sname}
            # sweep through lazy vals, eliminating those that depend on name; do this until none are
            # removed on a sweep
            go = True
            while go:
                tmp = [lnm for (lnm, (ldeps, lfn)) in self._lazy_vals.iteritems()
                           if lnm not in deps and any(dep in deps for dep in ldeps)]
                go = len(tmp) > 0
                deps.update(tmp)
            dep_tree[sname] = (deps - {sname})
        return dep_tree

    # This function will clear the lazily-evaluated members when a given value is changed
    def __update_values(self, name):
        for nm in self._lazy_deps[name]:
            if nm in self.__dict__:
                del self.__dict__[nm]

    # This is the most important function, given the encapsulation of this class:
    def __setattr__(self, name, val):
        if name in self._settable_vals:
            fn = self._settable_vals[name]
            self.__dict__[name] = fn(self, val)
            self.__update_values(name)
        elif name in self._lazy_vals:
            raise ValueError('The member %s is a lazy value and cannot be set' % name)
        else:
            raise ValueError('Unrecognized Immutable member: %s' % name)

    # The getattr method makes sure that lazy members are computed when requested
    def __getattr__(self, name):
        if name[0] == '_':
            return self.__dict__[name]
        if name in self.__dict__:
            return self.__dict__[name]
        elif name in self._lazy_vals:
            (deps, fn) = self._lazy_vals[name]
            tmp = fn(*[getattr(self, x) for x in deps])
            self.__dict__[name] = tmp
            return tmp
        elif name in self._const_vals:
            return self._const_vals[name];
        else:
            raise ValueError('Unrecognized member of Immutable: %s' % name)

    def __init__(self, settable_vals, const_vals, lazy_vals, **opts):
        self.__dict__['_settable_vals'] = make_dict(settable_vals).using(
            options=lambda s,o: make_dict(o))
        self.__dict__['_const_vals'] = make_dict(const_vals)
        self.__dict__['_lazy_vals'] = lazy_vals
        self.__dict__['_lazy_deps'] = self.__lazy_deps()
        self.options = opts

    def __repr__(self):
        return 'Immutable(<' + '>, <'.join([k for k,v in self._settable_vals.iteritems()]) + '>)'



#! /usr/bin/env python

from setuptools import setup

setup(
    name='neuropythy',
    version='0.1.1',
    description='Toolbox for flexible cortical mesh analysis and registration',
    keywords='neuroscience mesh cortex registration',
    author='Noah C. Benson',
    author_email='nben@nyu.edu',
    url='https://github.com/noahbenson/neuropythy/',
    license='GPLv3',
    packages=['neuropythy',
              'neuropythy.geometry', 
              'neuropythy.immutable',
              'neuropythy.topology',
              'neuropythy.freesurfer',
              'neuropythy.cortex',
              'neuropythy.registration'],
    package_data={'': ['LICENSE.txt', 'lib/nben/target/nben-standalone.jar']},
    install_requires=['numpy>=1.2',
                      'scipy>=0.7',
                      'nibabel>=2.0',
                      'pysistence>=0.4',
                      'py4j>=0.9'])



####################################################################################################
# mrvolume.py
# Representations of MRI volumes
# by Noah C. Benson

import numpy as np
import scipy as sp
import nibabel.freesurfer.io as fsio
import os




####################################################################################################
# neuropythy/freesurfer/subject.py
# Simple tools for use with FreeSurfer in Python
# By Noah C. Benson

import numpy as np
import numpy.linalg
import scipy as sp
import scipy.spatial as space
import nibabel.freesurfer.io as fsio
import nibabel.freesurfer.mghformat as fsmgh
import os, math
import itertools
import pysistence
from   pysistence import make_dict
from   neuropythy.cortex import CorticalMesh
from   neuropythy.topology import Topology
import neuropythy.geometry as geo

# These static functions are just handy
def spherical_distance(pt0, pt1):
    '''spherical_distance(a, b) yields the angular distance between points a and b, both of which
       should be expressed in spherical coordinates as (longitude, latitude).'''
    dtheta = pt1[0] - pt0[0]
    dphi   = pt1[1] - pt0[1]
    a = np.sin(dphi/2)**2 + np.cos(pt0[1]) * np.cos(pt1[1]) * np.sin(dtheta/2)**2
    return 2 * np.arcsin(np.sqrt(a))

class PropertyBox(list):
    '''PropertyBox is a simple class for lazily loading properties into a hemisphere or mesh. It
       should generally not be used directly and instead obtained from the Hemisphere class. Note
       that a PropertyBox object should behave like a list.'''
    def __init__(self, loader):
        self._loading_fn = loader
    def __len__(self):
        if 'data' not in self.__dict__: self.__load_data()
        return len(self.data)
    def __getitem__(self, key):
        if 'data' not in self.__dict__: self.__load_data()
        return self.data[key]
    def __iter__(self):
        if 'data' not in self.__dict__: self.__load_data()
        return iter(self.data)
    def __reversed__(self):
        if 'data' not in self.__dict__: self.__load_data()
        return reversed(self.data)
    def __contains__(self, item):
        if 'data' not in self.__dict__: self.__load_data()
        return item in self.data
    def __load_data(self):
        f = self._loading_fn
        dat = f()
        self.__dict__['data'] = dat

class Hemisphere:
    '''FreeSurfer.Hemisphere encapsulates the data contained in a subject's freesurfer hemisphere.
    This includes the various surface data as well as certain volume data.'''
    

    ################################################################################################
    # Lazy/Static Interface
    # This code handles the lazy interface; only three data are non-lazy and these are coordinates,
    # faces, and options; they may be set directly.

    # The following methods and static variable allow one to set particular members of the object
    # directly:
    @staticmethod
    def _check_property(self, name, val):
        if not isinstance(val, np.ndarray) and not isinstance(val, list):
            raise ValueError('property values must be lists or numpy arrays')
        return True
    @staticmethod
    def _check_properties(self, val):
        if not isinstance(val, dict):
            raise ValueError('properties must be a dictionary')
        for (k, v) in val.iteritems(): Hemisphere._check_property(self, k, v)
        if type(val) is pysistence.persistent_dict.PDict:
            return val
        else:
            return make_dict(**val)
    @staticmethod
    def _check_options(self, val):
        # Options just have to be a dictionary and are converted to an immutable one
        if not isinstance(val, dict):
            raise ValueError('options must be a dictionary')
        if type(val) is pysistence.persistent_dict.PDict:
            return val
        else:
            return make_dict(**val)
    __settable_members = {
        'properties': lambda m,v: Hemisphere._check_properties(m,v),
        'options': lambda m,v: Hemisphere._check_options(m,v)}

    # This static variable and these functions explain the dependency hierarchy in cached data
    def __make_surface(self, coords, faces, name, reg=None):
        return CorticalMesh(
            coords,
            faces,
            subject = self.subject,
            hemisphere = self,
            meta_data = self.meta_data.using(
                **{'subject': self.subject,
                   'hemisphere': self,
                   'name': name,
                   'registration': reg}))
    def _load_surface_data(self, name):
        path = self.subject.surface_path(name, self.name)
        if not os.path.exists(path):
            return None
        else:
            data = fsio.read_geometry(path)
            data[0].setflags(write=False)
            data[1].setflags(write=False)
            return data
    def _load_surface_data_safe(self, name):
        try:
            return self._load_surface_data(name)
        except:
            return (None, None)
    def _load_surface(self, name, preloaded=None, reg=None):
        data = self._load_surface_data(name) if preloaded is None else preloaded
        return self.__make_surface(data[0], data[1], name, reg=reg)
    def _load_sym_surface(self, name):
        path = self.subject.surface_path(name, self.name)
        if not os.path.exists(path):
            return None
        else:
            data = fsio.read_geometry(path)
            data[0].setflags(write=False)
            data[1].setflags(write=False)
            return self.__make_surface(data[0], data[1], name)
    def _load_ribbon(self):
        path = self.subject.volume_path('ribbon', self.name)
        if not os.path.exists(path):
            return None
        else:
            data = fsmgh.load(path)
            # for now, no post-processing, just loading of the MGHImage
            return data
    def _make_topology(self, sphere, fsave, fssym):
        if sphere is None:
            return None
        names = [self.subject.id, 'fsaverage', 'fsaverage_sym'];
        surfs = [sphere, fsave, fssym]
        return Topology(
            sphere.faces, 
            {names[i]: surfs[i].coordinates.T for i in range(len(names)) if surfs[i] is not None})
    @staticmethod
    def _check_meta_data(opts):
        md = opts.get('meta_data', {})
        if not isinstance(md, dict):
            raise ValueError('hemisphere meta-data must be a dictionary')
        if type(md) is pysistence.persistent_dict.PDict:
            return md
        else:
            return make_dict(**md)
        
    __lazy_members = {
        'meta_data':            (('options',), lambda hemi,opts: Hemisphere._check_meta_data(opts)),
        'white_surface':        ((), lambda hemi: hemi._load_surface('white')),
        'pial_surface':         ((), lambda hemi: hemi._load_surface('pial')),
        'inflated_surface':     ((), lambda hemi: hemi._load_surface('inflated')),

        'sphere_surface_data':  ((), lambda hemi: hemi._load_surface_data('sphere')),
        'fs_surface_data':      ((), lambda hemi: hemi._load_surface_data_safe('sphere.reg')),
        'sym_surface_data':     ((), 
                                 lambda hemi: \
                                    hemi._load_surface_data_safe('fsaverage_sym.sphere.reg')),
        'faces':                (('sphere_surface_data',), lambda hemi,dat: dat[1]),

        'sphere_surface':       (('sphere_surface_data','topology'), 
                                 lambda hemi,dat,topo: hemi._load_surface(
                                     'sphere',
                                     preloaded=dat,
                                     reg=(topo.registrations[hemi.subject.id]
                                          if hemi.subject.id in topo.registrations
                                          else None))),
        'fs_sphere_surface':    (('fs_surface_data','topology'),
                                 lambda hemi,dat,topo: (
                                    hemi.sphere_surface if hemi.subject.id == 'fsaverage'
                                    else None if dat == (None,None)
                                    else hemi._load_surface(
                                            'sphere.reg',
                                            preloaded=dat,
                                            reg=(topo.registrations['fsaverage']
                                                 if 'fsaverage' in topo.registrations
                                                 else None)))),
        'sym_sphere_surface':   (('sym_surface_data','topology'),
                                 lambda hemi,dat,topo: (
                                    hemi.sphere_surface if hemi.subject.id == 'fsaverage_sym'
                                    else None if dat is None
                                    else hemi._load_surface(
                                            'fsaverage_sym.sphere.reg',
                                            preloaded=dat,
                                            reg=(topo.registrations['fsaverage_sym']
                                                 if 'fsaverage_sym' in topo.registrations
                                                 else None)))),

        'vertex_count':         (('sphere_surface_data',), lambda hemi,dat: dat[0].shape[1]),
        'midgray_surface':      (('white_surface', 'pial_surface'),
                                 lambda hemi,W,P: hemi.__make_surface(
                                     0.5*(W.coordinates + P.coordinates),
                                     W.faces,
                                     'midgray')),
        'occipital_pole_index': (('inflated_surface',),
                                 lambda hemi,mesh: np.argmin(mesh.coordinates[1])),
        'ribbon':               ((), lambda hemi: hemi._load_ribbon()),
        'chirality':            (('name',), 
                                 lambda hemi,name: 'LH' if name == 'LH' or name == 'RHX' else 'RH'),
        'topology':             (('sphere_surface_data', 'fs_surface_data', 'sym_surface_data'),
                                 lambda hemi,sph,fs,sym: Topology(sph[1],
                                                                  {hemi.subject.id: sph[0],
                                                                   'fsaverage': fs[0],
                                                                   'fsaverage_sym': sym[0]}))}
    
    # This function will clear the lazily-evaluated members when a given value is changed
    def __update_values(self, name):
        for (sname, (deps, fn)) in Hemisphere.__lazy_members.items():
            if name in deps and sname in self.__dict__:
                del self.__dict__[sname]

    # This is the most important function, given the encapsulation of this class:
    def __setattr__(self, name, val):
        if name in Hemisphere.__settable_members:
            fn = Hemisphere.__settable_members[name]
            self.__dict__[name] = fn(self, val)
            self.__update_values(name)
        elif name in Hemisphere.__lazy_members:
            raise ValueError('The member %s is a lazy value and cannot be set' % name)
        else:
            raise ValueError('Unrecognized Hemisphere member: %s' % name)

    # The getattr method makes sure that lazy members are computed when requested
    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        elif name in Hemisphere.__lazy_members:
            (deps, fn) = Hemisphere.__lazy_members[name]
            tmp = fn(self, *map(lambda x: getattr(self, x), deps))
            self.__dict__[name] = tmp
            return tmp
        else:
            raise ValueError('Unrecognized member of Hemisphere: %s' % name)

    # Make a surface out of coordinates, if desired
    def surface(self, coords, name=None):
        '''
        hemi.surface(coords) yields a new surface with the given coordinate matrix and the topology
        of this hemisphere, hemi. The coordinate matrix must have hemi.vertex_count points.
        '''
        coords = np.asarray(coords)
        coords = (coords   if coords.shape[1] == self.vertex_count else
                  coords.T if coords.shape[0] == self.vertex_count else
                  None)
        if coords is None: raise ValueError('Coordinate matrix was invalid size!')
        return self.__make_surface(coords, self.faces, name)

    
    ################################################################################################
    # The Constructor
    def __init__(self, subject, type, **args):
        if not isinstance(subject, Subject):
            raise ValueError('Argument subject must be a FreeSurfer.Subject object')
        if type.upper() not in set(['RH', 'LH', 'RHX', 'LHX']):
            raise ValueError('Argument type must be RH, LH, RHX, or LHX')
        self.__dict__['subject'] = subject
        self.__dict__['name'] = type.upper()
        self.__dict__['directory'] = os.path.join(subject.directory, 'surf')
        self.properties = args.pop('properties', make_dict())
        self.options = args
        self.__init_properties()
    
    ################################################################################################
    # The display function
    def __repr__(self):
        return "Hemisphere(" + self.name + ", <subject: " + self.subject.id + ">)"

    
    ################################################################################################
    # Property code
    def add_property(self, name, prop=Ellipsis):
        '''hemi.add_property(name, prop) adds (or overwrites) the given property with the given name
           in the given hemi. The name must be a valid dictionary key and the prop argument must be
           a list of numpy array of values, one per vertex.
           hemi.add_property(d) adds all of the properties in the given dictionary of properties, 
           d.
           Note that in either case, if the value of prop or the value in a dictionary item is
           None, then the item is removed from the property list instead of being added.'''
        if prop is Ellipsis:
            if isinstance(name, dict):
                for (n,p) in name.iteritems():
                    self.add_property(n, p)
            else:
                raise ValueError('add_property must be called with a name and propery or a dict')
        else:
            if prop is None:
                self.remove_property(name)
            else:
                Hemisphere._check_property(self, name, prop)
                self.__dict__['properties'] = self.properties.using(**{name: prop})

    def property_names(self):
        '''hemi.property_names() yields a set of property names for the given hemi.'''
        return set(self.properties.keys())

    def has_property(self, name):
        '''hemi.has_property(name) yields True if the given hemisphere contains the property with
           the given name and False otherwise.'''
        return name in self.properties

    def remove_property(self, name):
        '''hemi.remove_property(name) removes the property with the given name from the given hemi.
           The name argument may also be an iterable collection of names, in which case all are
           removed.'''
        if hasattr(name, '__iter__'):
            for n in name: self.remove_property(n)
        elif name in self.properties:
            self.__dict__['properties'] = self.properties.without(name)
    
    def property_value(self, name):
        '''hemi.property_value(name) yields a list of the property values with the given name. If
           name is an iterable, then property_value automatically threads across it. If no such
           property is found in the mesh, then None is returned.'''
        if hasattr(name, '__iter__'):
            return map(lambda n: self.property_value(n), name)
        else:
            return self.properties.get(name, None)
            
    def prop(self, name, arg=Ellipsis):
        '''hemi.prop(...) is a generic function for handling mesh properties. It may be called
           with a variety of arguments:
             * hemi.prop(name) is equivalent to hemi.property_value(name) if name is either not
               an iterable or is an iterable that is not a dictionary
             * hemi.prop(d) is equivalent to hemi.add_property(d) if d is a dictionary
             * hemi.prop(name, val) is equivalent to hemi.add_property(name, val) if val is not
               None
             * hemi.prop(name, None) is equivalent to hemi.remove_property(name)'''
        if arg is Ellipsis:
            if isinstance(name, dict):
                self.add_property(name)
            else:
                return self.property_value(name)
        else:
            self.add_property(name, arg)

    def interpolate(self, from_hemi, property_name, 
                    apply=True, method='automatic', mask=None, null=None, n_jobs=1):
        '''
        hemi.interpolate(from_hemi, prop) yields a list of property values that have been resampled
        onto the given hemisphere hemi from the property with the given name prop of the given 
        from_mesh. If the optional apply is set to True (the default) or to a new property name
        string, the property is added to hemi as well; otherwise it is only returned. Optionally,
        prop may be a list of numpy array of values the same length as the mesh's vertex list; in
        this case, the apply option must be a new property name string, otherwise it is treated as
        False. Note that in order to work, the hemi and from_hemi objects must share a registration
        such as fsaverage or fsaverage_sym.

        Options:
          * mask (default: None) indicates that the given True/False or 0/1 valued list/array should
            be used; any point whose nearest neighbor (see below) is in the given mask will, instead
            of an interpolated value, be set to the null value (see null option).
          * null (default: None) indicates the value that should be placed in the returned result if
            either a vertex does not lie in any triangle or a vertex is masked out via the mask
            option.
          * smoothing (default: 2) assuming that the method is 'interpolate' or 'automatic', this
            is the exponent used to smooth the interpolated surface; 1 is pure linear interpolation
            while 2 represents a slightly smoother version of this. Note that this is not an order
            of interpolation option.
          * method (default: 'automatic') specifies what method to use for interpolation. The only
            currently supported methods are 'automatic' or 'nearest'. The 'nearest' method does not
            actually perform a nearest-neighbor interpolation but rather assigns to a destination
            vertex the value of the source vertex whose veronoi-like polygon contains the
            destination vertex; note that the term 'veronoi-like' is used here because it uses the
            Voronoi diagram that corresponds to the triangle mesh and not the true delaunay
            triangulation. The 'automatic' checks every destination vertex and assigns it the
            'nearest' value if that value would not be a number, otherwise it interpolates linearly
            within the vertex's source triangle.
          * n_jobs (default: 1) is passed along to the cKDTree.query method, so may be set to an
            integer to specify how many processors to use, or may be -1 to specify all processors.
        '''
        if from_hemi.chirality != self.chirality:
            raise ValueError('hemispheres have opposite chiralities')
        if isinstance(property_name, basestring):
            if not from_hemi.has_property(property_name):
                raise ValueError('given property ' + property_name + ' is not in from_hemi!')
            data = from_hemi.prop(property_name)
        elif type(property_name).__module__ == np.__name__:
            data = property_name
            property_name = apply if isinstance(apply, basestring) else None
        elif hasattr(property_name, '__iter__'):
            if all(isinstance(s, basestring) for s in property_name):
                return {p: self.interpolate(from_hemi, p, 
                                            method=method, mask=mask,
                                            null=null, n_jobs=n_jobs)
                        for p in property_name}
            else:
                data = np.asarray(property_name)
                property_name = apply if isinstance(apply, basestring) else None
        else:
            raise ValueError('property_name is not a string or valid list')
        # pass data along to the topology object...
        result = self.topology.interpolate_from(from_hemi.topology, data,
                                                method=method, mask=mask,
                                                null=null, n_jobs=n_jobs)
        if result is not None:
            if apply is True and property_name is not None:
                self.prop(property_name, result)
            elif isinstance(apply, basestring):
                self.prop(apply, result)
        return result

    # This [private] function and this variable set up automatic properties from the FS directory
    # in order to be auto-loaded, a property must appear in this dictionary:
    _auto_properties = {
        'sulc':      ('convexity',              lambda f: fsio.read_morph_data(f)),
        'thickness': ('thickness',              lambda f: fsio.read_morph_data(f)),
        'area':      ('white_surface_area',     lambda f: fsio.read_morph_data(f)),
        'area.mid':  ('midgray_surface_area',   lambda f: fsio.read_morph_data(f)),
        'area.pial': ('pial_surface_area',      lambda f: fsio.read_morph_data(f)),
        'curv':      ('curvature',              lambda f: fsio.read_morph_data(f)),

        'prf_eccen': ('PRF_eccentricity',       lambda f: fsio.read_morph_data(f)),
        'prf_angle': ('PRF_polar_angle',        lambda f: fsio.read_morph_data(f)),
        'prf_size':  ('PRF_size',               lambda f: fsio.read_morph_data(f)),
        'prf_varex': ('PRF_variance_explained', lambda f: fsio.read_morph_data(f)),

        'retinotopy_eccen': ('eccentricity',    lambda f: fsio.read_morph_data(f)),
        'retinotopy_angle': ('polar_angle',     lambda f: fsio.read_morph_data(f)),
        'retinotopy_areas': ('visual_area',     lambda f: fsio.read_morph_data(f))}
    # properties grabbed out of MGH or MGZ files in the sufsio.read_morph_data(f)),
    _mgh_properties = {}
    # funciton for initializing the auto-loading properties
    def __init_properties(self):
        dir = self.directory
        files = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]            
        autoprops = Hemisphere._auto_properties
        for file in files:
            if len(file) > 2 and file[0:2].upper() == self.chirality and file[3:] in autoprops:
                (name, fn) = autoprops[file[3:]]
                #self.prop(name, PropertyBox(lambda: fn(os.path.join(dir, file))))
                self.prop(name, fn(os.path.join(dir, file)))
        # We also want to auto-add labels:
        dir = os.path.join(self.subject.directory, 'label')
        files = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]
        for file in files:
            if len(file) > 9 and file[0:2].upper() == self.chirality and file[-6:] == '.label':
                if len(file) < 17 or file[-13:-6] != '.thresh':
                    lbl = set(fsio.read_label(os.path.join(dir, file)))
                    self.prop(
                        file[3:-6],
                        [True if k in lbl else False for k in range(self.vertex_count)])
                #else:
                #    (lbl, sclr) = fsio.read_label(os.path.join(dir, file), read_scalars=True)
                #    lbl = {lbl[i]: i for i in range(len(lbl))}
                #    self.prop(
                #        file[3:-13] + '_threshold',
                #        [sclr[lbl[k]] if k in lbl else None for k in range(self.vertex_count)])
        # And MGH/MGZ files in surf:
        dir = os.path.join(self.subject.directory, 'surf')
        files = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]
        for f in files:
            if (f[-4:].lower() == '.mgh' or f[-4:].lower() == '.mgz') and \
                    f[2] == '.' and f[0:2].upper() == self.chirality and \
                    f[3:-4] in Hemisphere._mgh_properties:
                (name, fn) = Hemisphere._mgh_properties[f[3:-4]]
                self.prop(name, fn(os.path.join(dir, f)))

    # This method is a convenient way to get the occipital pole coordinates for the various
    # surfaces in a hemisphere...
    def occipital_pole(self, surfname):
        surfname = surfname.lower()
        if not surfname.lower().endswith('_surface'):
            surfname = surfname + '_surface'
        surf = self.__getattr__(surfname)
        opIdx = self.occipital_pole_index
        return surf.coordinates[:, opIdx]

    # These methods, values, and the projection fn are all related to map projection
    @staticmethod
    def _orthographic_projection(X, params): 
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        return X[1:3]
    @staticmethod
    def _orthographic_projection_inverse(X, params): 
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        r = params['sphere_radius']
        Xnorm = X / r
        return np.asarray([r * np.sqrt(1.0 - (Xnorm ** 2).sum(0)), X[0], X[1]])
    @staticmethod
    def _equirectangular_projection(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        X = X / np.sqrt((X ** 2).sum(0))
        r = params['sphere_radius']
        return r / math.pi * np.asarray([np.arctan2(X[1], X[0]), np.arcsin(X[2])])
    @staticmethod
    def _equirectangular_projection_inverse(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        r = params['sphere_radius']
        X = math.pi * X / r
        return np.asarray([np.cos(X[0]) * r, 
                           np.sin(X[0]) * r,
                           np.sin(X[1]) * r])
    @staticmethod
    def _mercator_projection(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        X = X / np.sqrt((X ** 2).sum(0))
        r = params['sphere_radius']
        return r * np.asarray([np.arctan2(X[1], X[0]),
                               np.log(np.tan(0.25 * math.pi + 0.5 * np.arcsin(X[2])))])
    @staticmethod
    def _mercator_projection_inverse(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        r = params['sphere_radius']
        X = X / r
        return r * np.asarray([np.cos(X[0]),
                               np.sin(X[0]),
                               np.sin(2 * (np.arctan(np.exp(X[1])) - 0.25*math.pi))])
    @staticmethod
    def _sinusoidal_projection(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        X = X / np.sqrt((X ** 2).sum(0))
        r = params['sphere_radius']
        phi = np.arcsin(X[2])
        return r / math.pi * np.asarray([np.arctan2(X[1], X[0]) * np.cos(phi), phi])
    @staticmethod
    def _sinusoidal_projection_inverse(X, params):
        X = np.asarray(X)
        X = X if X.shape[0] < 4 else X.T
        r = params['sphere_radius']
        X = math.pi * X / r
        z = np.sin(X[1])
        cosphi = np.cos(X[1])
        return np.asarray([np.cos(X[0] / cosphi) * r,
                           np.sin(X[0] / cosphi) * r,
                           np.sin(X[1]) * r])

    __projection_methods = {
        'orthographic':    lambda X,p: Hemisphere._orthographic_projection(X,p),
        'equirectangular': lambda X,p: Hemisphere._equirectangular_projection(X,p),
        #'mollweide':       lambda X,p: Hemisphere._mollweide_projection(X,p),
        'mercator':        lambda X,p: Hemisphere._mercator_projection(X,p),
        'sinusoidal':      lambda X,p: Hemisphere._sinusoidal_projection(X,p)}
    __projection_methods_inverse = {
        'orthographic':    lambda X,p: Hemisphere._orthographic_projection_inverse(X,p),
        'equirectangular': lambda X,p: Hemisphere._equirectangular_projection_inverse(X,p),
        #'mollweide':       lambda X,p: None,
        'mercator':        lambda X,p: Hemisphere._mercator_projection_inverse(X,p),
        'sinusoidal':      lambda X,p: Hemisphere._sinusoidal_projection_inverse(X,p)}
    def __interpret_projection_params(self, params):
        # Figure out which spherical surface we're using...
        if 'surface' not in params or params['surface'].lower() == 'native':
            params['surface'] = 'native'
            params['mesh'] = self.sphere_surface
        elif params['surface'].lower() == 'fsaverage':
            params['surface'] = 'fsaverage'
            params['mesh'] = self.fs_sphere_surface
        elif params['surface'].lower() == 'fsaverage_sym':
            params['surface'] = 'fsaverage_sym'
            params['mesh'] = self.sym_sphere_surface
        elif params['surface'] in self.topology.registrations:
            surfname = params['surface']
            params['mesh'] = self.__make_surface(
                self.topology.registrations[surfname].coordinates.T,
                self.faces,
                surfname)
        else:
            raise ValueError('Unrecognized spherical surface: %s' % params['surface'])
        sphere = params['mesh']
        # important parameters: center and radius => indices
        if 'center' not in params:
            params['center'] = sphere.coordinates[:, self.occipital_pole_index]
        elif isinstance(params['center'], int):
            params['center'] = sphere.coordinates[:, params['center']]
        if 'radius' not in params:
            params['radius'] = math.pi / 4.0
        elif not isinstance(params['radius'], (int, long, float)) or params['radius'] <= 0:
            raise ValueError('radius parameter must be a positive integer')
        # also, we need to worry about the method...
        if 'method' not in params:
            params['method'] = 'equirectangular'
        elif params['method'].lower() not in Hemisphere.__projection_methods:
            raise ValueError('method given to projection not recognized: %s' % params['method'])
        # Finally, we need to look at exclusions...
        # (#TODO)
        return params
    @staticmethod
    def __select_projection_subset(mesh, center, radius):
        # Figure out which vertices and faces to include, yield the indices of each
        # use the spherical coordinates...
        center = np.asarray(center)
        sc = mesh.spherical_coordinates
        center_sc = [np.arctan2(center[0], center[1]),
                     np.arcsin(center[2] / np.sqrt(sum(center ** 2)))]
        return mesh.select(lambda u: spherical_distance(center_sc, sc[0:2,u]) < radius)
    def projection(self, **params):
        params = self.__interpret_projection_params(params)
        submesh = Hemisphere.__select_projection_subset(
            params['mesh'],
            params['center'],
            params['radius'])
        params['sphere_radius'] = np.mean(np.sqrt((submesh.coordinates ** 2).sum(0)))
        FR = geo.alignment_matrix_3D(params['center'], [1.0, 0.0, 0.0])
        IR = geo.alignment_matrix_3D([1.0, 0.0, 0.0], params['center'])
        fwdprojfn = Hemisphere.__projection_methods[params['method']]
        invprojfn = Hemisphere.__projection_methods_inverse[params['method']]
        fwdfn = lambda X,p: fwdprojfn(np.dot(FR, X), p)
        invfn = lambda X,p: np.dot(IR, invprojfn(X, p))
        params['forward_function'] = fwdfn
        params['inverse_function'] = invfn
        params = make_dict(params)
        submesh.coordinates = fwdfn(submesh.coordinates, params)
        submesh.options = submesh.options.using(projection_parameters=params)
        for p in self.property_names():
            try:
                submesh.prop(p, np.asarray(self.prop(p))[submesh.vertex_labels])
            except:
                pass
        return submesh

class Subject:
    '''FreeSurfer.Subject objects encapsulate the data contained in a FreeSurfer
       subject directory.'''

    ################################################################################################
    # Lazy/Static Interface
    # This code handles the lazy interface; only three data are non-lazy and these are coordinates,
    # faces, and options; they may be set directly.

    # The following methods and static variable allow one to set particular members of the object
    # directly:
    @staticmethod
    def _check_options(self, val):
        # Options just have to be a dictionary and are converted to an immutable one
        if not isinstance(val, dict):
            raise ValueError('options must be a dictionary')
        if type(val) is pysistence.persistent_dict.PDict:
            return val
        else:
            return make_dict(**val)
    __settable_members = {
        'options': lambda m,v: Subject._check_options(m,v)}

    # This static variable and these functions explain the dependency hierarchy in cached data
    @staticmethod
    def _check_meta_data(opts):
        md = opts.get('meta_data', {})
        if not isinstance(md, dict):
            raise ValueError('subject meta-data must be a dictionary')
        if type(md) is pysistence.persistent_dict.PDict:
            return md
        else:
            return make_dict(**md)
    def _load_hemisphere(self, name):
        return Hemisphere(self, name)
    __lazy_members = {
        'meta_data': (('options',), lambda mesh,opts: Subject._check_meta_data(opts)),
        'LH':  ((), lambda mesh: mesh._load_hemisphere('LH')),
        'RH':  ((), lambda mesh: mesh._load_hemisphere('RH')),
        'LHX': ((), lambda mesh: mesh._load_hemisphere('LHX')),
        'RHX': ((), lambda mesh: mesh._load_hemisphere('RHX'))}
    
    # This function will clear the lazily-evaluated members when a given value is changed
    def __update_values(self, name):
        for (sname, (deps, fn)) in Subject.__lazy_members.items():
            if name in deps and sname in self.__dict__:
                del self.__dict__[sname]

    # This is the most important function, given the encapsulation of this class:
    def __setattr__(self, name, val):
        if name in Subject.__settable_members:
            fn = Subject.__settable_members[name]
            self.__dict__[name] = fn(self, val)
            self.__update_values(name)
        elif name in Subject.__lazy_members:
            raise ValueError('The member %s is a lazy value and cannot be set' % name)
        else:
            raise ValueError('Unrecognized Subject member: %s' % name)

    # The getattr method makes sure that lazy members are computed when requested
    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        elif name in Subject.__lazy_members:
            (deps, fn) = Subject.__lazy_members[name]
            tmp = fn(self, *map(lambda x: getattr(self, x), deps))
            self.__dict__[name] = tmp
            return tmp
        else:
            raise ValueError('Unrecognized member of Subject: %s' % name)

    
    ################################################################################################
    # The Constructor
    def __init__(self, subject, **args):
        (dir, name) = os.path.split(subject)
        if not os.path.isdir(subject):
            if dir is '':
                dir = args.pop('subjects_dir', os.environ['SUBJECTS_DIR'])
        elif dir is '':
            dir = '.'
        subpath = os.path.join(dir, name)
        if not os.path.isdir(subpath):
            raise ValueError('Subject directory not found: %s' % subpath)
        if not os.path.isdir(os.path.join(subpath, 'surf')):
            raise ValueError('Subject surf directory not found')
        self.__dict__['subjects_dir'] = dir
        self.__dict__['id'] = name
        self.__dict__['directory'] = subpath
        self.options = args
    
    ################################################################################################
    # The display function
    def __repr__(self):
        return "Subject(" + self.id + ")"


    ################################################################################################
    # Standard Methods

    def surface_path(self, type, hemi):
        hemi = hemi.upper()
        if hemi == 'RHX':
            if not os.path.isdir(os.path.join(self.directory, 'xhemi', 'surf')):
                raise ValueError('Subject does not have an xhemi/surf directory')
            path = os.path.join(self.directory, 'xhemi', 'surf', 'lh.' + type)
        elif hemi == 'LHX':
            if not os.path.isdir(os.path.join(self.directory, 'xhemi', 'surf')):
                raise ValueError('Subject does not have an xhemi/surf directory')
            path = os.path.join(self.directory, 'xhemi', 'surf', 'rh.' + type)
        elif hemi == 'LH':
            path = os.path.join(self.directory, 'surf', 'lh.' + type)
        elif hemi == 'RH':
            path = os.path.join(self.directory, 'surf', 'rh.' + type)
        else:
            raise ValueError('Argument hemi must be LH, RH, LHX, or RHX')
        if not os.path.isfile(path):
            raise ValueError('file %s not found' % path)
        return path
    def volume_path(self, type, hemi):
        # hemi may be None, in which case we have only these surfaces:
        if hemi is None:
            nm = type + '.'
        else:
            nm = hemi.lower() + '.' + type + '.'
        flnm = os.path.join(self.directory, 'mri', nm + 'mgz')
        if not os.path.isfile(flnm):
            flnm = os.path.join(self.directory, 'mri', nm + 'mgh')
            if not os.path.isfile(flnm):
                raise ValueError('no such file (or mgz equivalent) found: ' + flnm)
        return flnm
            

####################################################################################################
# Some FreeSurfer specific functions

def cortex_to_ribbon_map_smooth(sub, hemi=None, k=12, distance=4, sigma=0.35355):
    '''cortex_to_ribbon_map_smooth(sub) yields a dictionary whose keys are the indices of the ribbon
         voxels for the given FreeSurfer subject sub and whose values are a tuple of both (0) a list
         of vertex labels associated with the given voxel and (1) a list of the weights associated
         with each vertex (both in identical order).

       These associations are determined by finding the k nearest neighbors (of the midgray surface)
       to each voxel then finding the lines from white-to-pial that pass within a certain distance
       of the voxel center and weighting each appropriately.

       The following options are accepted:
         * k (default: 4) is the number of vertices to find near each voxel center
         * distance (default: 3) is the cutoff for including a vertex in a voxel region
         * hemi (default: None) specifies which hemisphere to operate over; this may be 'lh', 'rh',
           or None (to do both)
         * sigma (default: 1 / (2 sqrt(2))) specifies the standard deviation of the Gaussian used
           to weight the vertices contributing to a voxel'''
    
    # we can speed things up slightly by doing left and right hemispheres separately
    if hemi is None:
        return (cortex_to_ribbon_map(sub, k=k, distance=distance, sigma=sigma, hemi='lh'),
                cortex_to_ribbon_map(sub, k=k, distance=distance, sigma=sigma, hemi='rh'))
    if not isinstance(hemi, basestring):
        raise ValueError('hemi must be a string \'lh\', \'rh\', or None')
    if hemi.lower() == 'lh' or hemi.lower() == 'left':
        hemi = sub.LH
    elif hemi.lower() == 'rh' or hemi.lower() == 'right':
        hemi = sub.RH

    # first we find the indices of all voxels that are in the ribbon
    ribdat = hemi.ribbon.get_data()
    idcs = np.transpose(ribdat.nonzero()) + 1  # voxel transforms assume 1-based indexing
    # we also need the transformation from surface to voxel
    #tmtx = hemi.ribbon.header.get_ras2vox()
    tmtx = np.linalg.inv(hemi.ribbon.header.get_vox2ras_tkr())

    # now we want to make an octree from the midgray voxels
    txcoord = lambda mtx: np.dot(tmtx[0:3], np.vstack((mtx, np.ones(mtx.shape[1]))))
    midgray = txcoord(hemi.midgray_surface.coordinates)
    near = space.KDTree(midgray.T)
    
    # then we find the k nearest vertices for each voxel with a distance cutoff
    (d, nei) = near.query(idcs, k=k, distance_upper_bound=distance, p=2)
    ## we want the i'th row of d and nei to be a list of the i'th-nearests to each voxel center
    d = d.T
    nei = nei.T
    
    # okay, now we want to find the closest point along each line segment... we do this iteratively;
    # in the process we replace the d array with the distance of the nearest point along the segment
    # between pial and white for the given vertex. Any inf value will be left alone and will become
    # a zero when run through exp

    ## grab the pial and white matter coordinates...
    pialX = txcoord(hemi.pial_surface.coordinates)
    whiteX = txcoord(hemi.white_surface.coordinates)
    pial2white = whiteX - pialX
    pial2white_norm = np.sqrt(np.sum(pial2white ** 2, 0))
    same_idcs = np.where(pial2white_norm < 1e-9)
    pial2white_norm[same_idcs] = 1
    pial2white[:, same_idcs] = 0
    pial2white_normed = pial2white / pial2white_norm
    ## At this point we also want a transpose of indices
    idcs = idcs.T
    ## Iterate over the neighbors
    for i in range(d.shape[0]):
        # which elements are finite? (which do we care about)
        (finite_vox_idcs,) = np.where(d[i] != np.inf)
        # we do a closest-point search along the segment from white to pial in these locations
        ## to start, we project out the normal
        finite_vtx_idcs = nei[i, finite_vox_idcs]
        finite_vox_inplane = idcs[:, finite_vox_idcs] - pialX[:, finite_vtx_idcs]
        finite_p2w_normed = pial2white_normed[:, finite_vtx_idcs]
        finite_vox_prods = np.sum(finite_vox_inplane * finite_p2w_normed, 0)
        finite_vox_ipnorm = np.sqrt(np.sum(finite_vox_inplane ** 2, 0))
        ## If the product is less than 0, that means the voxel center is outside of the pial surface
        (inval,) = np.where((finite_vox_prods < -0.1 * finite_vox_ipnorm) \
                                + (finite_vox_prods > 1.1 * finite_vox_ipnorm))
        if len(inval) > 0:
            # annotate that these are invalid globally
            d[i, finite_vox_idcs[inval]] = np.inf
            # and locally
            finite_vox_idcs = np.delete(finite_vox_idcs, inval)
            finite_vox_inplane = np.delete(finite_vox_inplane, inval, 1)
            finite_p2w_normed = np.delete(finite_p2w_normed, inval, 1)
            finite_vox_prods = np.delete(finite_vox_prods, inval)
        ## otherwise, find the distance along the plane to the voxel center...
        finite_vox_inplane -= finite_p2w_normed * finite_vox_prods
        ## the distance to the line is just the norm of this then
        mindist2s = np.sum(finite_vox_inplane ** 2, 0)
        # We now have the in-plane distances to the closest vertices; we can just save this
        d[i, finite_vox_idcs] = mindist2s
    ## At this point we've filled the d-values (those not infinite) to the in-plane squared-
    ## distances from the relevant voxel center to the surface segment; we can turn d into weights;
    ## all infinite distances will automatically get a weight of 0
    d = np.exp(-0.5 * d / (sigma * sigma))
    ## we want to divide weights by the total weight so that all weight columns add up to 1...
    wtot = np.sum(d, 0)
    ### we have to check if any of these values are 0 and give them a nearest-vertex assignment
    zero_idcs = np.where(wtot < 1e-9)
    if len(zero_idcs) > 0:
        (d0s, nei0s) = near.query(idcs.T[zero_idcs], k=1, p=2)
        nei[0, zero_idcs] = nei0s
        d[0, zero_idcs] = 1.0
        d[1:, zero_idcs] = 0.0
        wtot[zero_idcs] = 1.0
    ### okay, now divide
    d /= wtot
    
    # With d = weights and nei = indices, we can now build the dictionary
    vlab = np.array(hemi.white_surface.vertex_labels)
    res = {}
    for i in range(idcs.shape[1]):
        (wh,) = np.where(d[:,i] > 0)
        res[tuple(idcs[:,i].tolist())] = (
            vlab[nei[wh, i]].tolist(),
            d[wh, i].tolist())
    return res

def _line_voxel_overlap(vx, xs, xe):
    '''
    _line_voxel_overlap((i,j,k), xs, xe) yields the fraction of the vector from xs to xe that
      overlaps with the voxel (i,j,k); i.e., the voxel that starts at (i,j,k) and ends at
      (i,j,k) + 1.
    '''
    # We want to consider the line segment from smaller to larger coordinate:
    seg = [(x0,x1) if x0 <= x1 else (x1,x0) for (x0,x1) in zip(xs,xe)]
    # First, figure out intersections of start and end values
    isect = [(min(seg_end, vox_end), max(seg_start, vox_start))
             if seg_end >= vox_start and seg_start <= vox_end
             else None
             for ((seg_start, seg_end), (vox_start, vox_end)) in zip(seg, [(x,x+1) for x in vx])]
    # If any of these are None, we can't possibly overlap
    if None in isect or any(a == b for (a,b) in isect):
        return 0.0
    return np.linalg.norm([e - s for (s,e) in isect]) / np.linalg.norm([e - s for (s,e) in zip(xs,xe)])

def cortex_to_ribbon_map_lines(sub, hemi=None):
    '''
    cortex_to_ribbon_map_lines(sub) yields a dictionary whose keys are the indices of the ribbon 
      voxels for the given FreeSurfer subject sub and whose values are a tuple of both (0) a list of
      vertex labels associated with the given voxel and (1) a list of the weights associated with
      each vertex (both in identical order).

    These associations are determined by projecting the vectors from the white surface vertices to 
      the pial surface vertices into the the ribbon and weighting them by the fraction of the vector 
      that lies in the voxel.

    The following options are accepted:
      * hemi (default: None) specifies which hemisphere to operate over; this may be 'lh', 'rh', or
        None (to do both)
    '''
    
    # we can speed things up slightly by doing left and right hemispheres separately
    if hemi is None:
        return (cortex_to_ribbon_map_lines(sub, hemi='lh'),
                cortex_to_ribbon_map_lines(sub, hemi='rh'))
    if not isinstance(hemi, basestring):
        raise ValueError('hemi must be a string \'lh\', \'rh\', or None')
    if hemi.lower() == 'lh' or hemi.lower() == 'left':
        hemi = sub.LH
    elif hemi.lower() == 'rh' or hemi.lower() == 'right':
        hemi = sub.RH
    else:
        raise RuntimeError('Unrecognized hemisphere: ' + hemi)

    # first we find the indices of all voxels that are in the ribbon
    ribdat = hemi.ribbon.get_data()
    # 1-based indexing is assumed:
    idcs = map(tuple, np.transpose(ribdat.nonzero()) + 1)
    # we also need the transformation from surface to voxel
    tmtx = np.linalg.inv(hemi.ribbon.header.get_vox2ras_tkr())
    # given that the voxels assume 1-based indexing, this means that the center of each voxel is at
    # (i,j,k) for integer (i,j,k) > 0; we want the voxels to start and end at integer values with 
    # respect to the vertex positions, so we subtract 1/2 from the vertex positions, which should
    # make the range 0-1, for example, cover the vertices in the first voxel
    txcoord = lambda mtx: np.dot(tmtx[0:3], np.vstack((mtx, np.ones(mtx.shape[1])))) + 1.5

    # Okay; get the transformed coordinates for white and pial surfaces:
    pialX = txcoord(hemi.pial_surface.coordinates)
    whiteX = txcoord(hemi.white_surface.coordinates)
    
    # make a list of voxels through which each vector passes:
    min_idx = [min(i) for i in idcs]
    max_idx = [max(i) for i in idcs]
    vtx_voxels = [((i,j,k), (id, olap))
                  for (id, (xs, xe)) in enumerate(zip(whiteX.T, pialX.T))
                  for i in range(int(math.floor(min(xs[0], xe[0]))), int(math.ceil(max(xs[0], xe[0]))))
                  for j in range(int(math.floor(min(xs[1], xe[1]))), int(math.ceil(max(xs[1], xe[1]))))
                  for k in range(int(math.floor(min(xs[2], xe[2]))), int(math.ceil(max(xs[2], xe[2]))))
                  for olap in [_line_voxel_overlap((i,j,k), xs, xe)]
                  if olap > 0]
    
    # and accumulate these lists... first group by voxel index then sum across these
    first_fn = lambda x: x[0]
    vox_byidx = {
        vox: ([q[0] for q in dat], [q[1] for q in dat])
        for (vox,xdat) in itertools.groupby(sorted(vtx_voxels, key=first_fn), key=first_fn)
        for dat in [[q[1] for q in xdat]]}
    return {
        idx: (ids, np.array(olaps) / np.sum(olaps))
        for idx in idcs
        if idx in vox_byidx
        for (ids, olaps) in [vox_byidx[idx]]}

def cortex_to_ribbon_map(sub, hemi=None, method='lines', options={}):
    '''
    cortex_to_ribbon_map(sub) yields a dictionary whose keys are the indices of the ribbon voxels
      for the given FreeSurfer subject sub and whose values are a tuple of both (0) a list of vertex
      labels associated with the given voxel and (1) a list of the weights associated with each
      vertex (both in identical order).

    The following options may be given:
      * hemi (default: None) may be 'lh', 'rh', or None; if Nonem then the result is a tuple of the
        (LH_dict, RH_dict) where LH_dict and RH_dict are the individual results from calling the
        cortex_to_ribbon_map function with hemi set to 'lh' or 'rh', respectively
      * method (default: 'lines') may be 'lines' or 'smooth'; the lines method projects the line
        segments from the white to the pial into the volume and weights each voxel by the length of
        the line segments contained in them; the smooth method uses Gaussian weights based on
        distance of the nearby vertices to the voxel center; see cortex_to_ribbon_map_lines and
        cortex_to_ribbon_map_smooth for more details.
      * options (default: {}) may be set to a dictionary of options to pass along to the lines or
        smooth functions
    '''
    if method.lower() == 'lines':
        return cortex_to_ribbon_map_lines(sub, hemi=hemi, **options)
    elif method.lower() == 'smooth':
        return cortex_to_ribbon_map_smooth(sub, hemi=hemi, **options)
    else:
        raise RuntimeError('Unrecognized method: ' + str(method))


def _cortex_to_ribbon_map_into_volume_array(vol, m, dat):
    dat = np.array(dat)
    for k,v in m.items():
        vol[k[0]-1, k[1]-1, k[2]-1] = np.dot(dat[v[0]], v[1])
    return vol

def cortex_to_ribbon(sub, data, map=None, k=12, distance=6, hemi=None, sigma=0.35355, 
                     default=0, dtype=np.float32):
    '''cortex_to_ribbon(sub, data, args...) applies the cortical-surface to ribbon transformation
         that is represented by the result of cortex_to_ribbon_map(sub, args...). The option 'map'
         may also be given if the cortex_to_ribbon_map calls have already been made, and the option
         'default' may be used to specify what value non-ribbon voxels will take.'''
    hemi = hemi.lower() if isinstance(hemi, basestring) else hemi
    if map is None:
        map = cortex_to_ribbon_map(sub, k=k, distance=distance, hemi=hemi, sigma=sigma)
    if hemi is None:
        if not isinstance(map, tuple) or len(map) != 2:
            raise ValueError('map must match hemi argument')
        if not isinstance(data, tuple) or len(data) != 2:
            raise ValueError('data must match hemi argument')
    if not isinstance(map, tuple):
        map = (map, None) if hemi == 'lh' else (None, map)
    if not isinstance(data, tuple):
        data = (data, None) if hemi == 'lh' else (None, data)
    # start with a duplicate of the left ribbon data
    vol0 = sub.LH.ribbon if hemi is None or hemi == 'lh' else sub.RH.ribbon
    vol0dims = vol0.get_data().shape
    arr = default * np.ones(vol0dims) if default != 0 else np.zeros(vol0dims)
    # apply the surface data maps to the volume
    if hemi is None or hemi == 'lh':
        arr = _cortex_to_ribbon_map_into_volume_array(arr, map[0], data[0])
    if hemi is None or hemi == 'rh':
        arr = _cortex_to_ribbon_map_into_volume_array(arr, map[1], data[1])
    hdr = vol0.header.copy()
    hdr.set_data_dtype(dtype)
    return fsmgh.MGHImage(arr, vol0.affine, hdr, vol0.extra, vol0.file_map)
    
    
    
    
    
